"""
Tamr - Golden Records
See https://docs.tamr.com/docs/overview-golden-records
"""
from tamr_client.golden_records import project
from tamr_client.golden_records._golden_records import (
    _publish_async,
    _update_async,
    publish,
    update,
)



from abc import ABC as AbstractBaseClass


class BaseResource(AbstractBaseClass):
    """Base class for client-side resources.

    :param :class:`~tamr_unify_client.Client` client: Delegate underlying API calls to this client.
    :param :py:class:`dict` data: JSON data received from server for this resource
    :param str alias: API alias for this resource.
    | If set to ``None``, the ``api_path`` of this instance will be set to ``data["relativeId"]``.
    | Default: ``None``.
    """

    def __init__(self, client, data, alias=None):
        self.client = client
        self._data = data or {}
        self.api_path = alias or data.get("relativeId")

    @classmethod
    def from_data(cls, client, data, api_path=None):
        return cls(client, data, api_path)

    @property
    def relative_id(self):
        """:type: str"""
        return self._data.get("relativeId")

    @property
    def resource_id(self):
        """:type: str"""
        rid = self.relative_id
        if rid is None:
            return None
        return rid.split("/")[-1]

    def delete(self):
        """Deletes this resource. Some resources do not support deletion, and will raise a 405 error if this is called.

        :return: HTTP response from the server
        :rtype: :class:`requests.Response`
        """
        response = self.client.delete(self.api_path).successful()
        return response



from tamr_unify_client.base_resource import BaseResource
from tamr_unify_client.operation import Operation


class MachineLearningModel(BaseResource):
    """A Tamr Machine Learning model."""

    @classmethod
    def from_json(cls, client, resource_json, api_path=None):
        return super().from_data(client, resource_json, api_path)

    def train(self, **options):
        """Learn from verified labels.

        :param ``**options``: Options passed to underlying :class:`~tamr_unify_client.operation.Operation` .
            See :func:`~tamr_unify_client.operation.Operation.apply_options` .
        :returns: The resultant operation.
        :rtype: :class:`~tamr_unify_client.operation.Operation`
        """
        op_json = self.client.post(self.api_path + ":refresh").successful().json()
        op = Operation.from_json(self.client, op_json)
        return op.apply_options(**options)

    def predict(self, **options):
        """Suggest labels for unverified records.

        :param ``**options``: Options passed to underlying :class:`~tamr_unify_client.operation.Operation` .
            See :func:`~tamr_unify_client.operation.Operation.apply_options` .
        :returns: The resultant operation.
        :rtype: :class:`~tamr_unify_client.operation.Operation`
        """
        dependent_dataset = "/".join(self.api_path.split("/")[:-1])
        op_json = self.client.post(dependent_dataset + ":refresh").successful().json()
        op = Operation.from_json(self.client, op_json)
        return op.apply_options(**options)

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"relative_id={self.relative_id!r})"
        )



from tamr_client.exception import TamrClientException


class Ambiguous(TamrClientException):
    """Raised when referencing a primary key by name that matches multiple possible targets."""

    pass


class NotFound(TamrClientException):
    """Raised when referencing a primary key by name that does not exist."""

    pass



"""Adaptation of the Python standard library JSONEncoder to encode `NaN` as 'null'
Compare to https://github.com/python/cpython/blob/3.9/Lib/json/encoder.py
The only functional difference is in the definition of `floatstr` where 'NaN', 'Infinity', and '-Infinity' are encoded as 'null'
"""
from json import JSONEncoder
from json.encoder import (
    _make_iterencode,
    py_encode_basestring,
    py_encode_basestring_ascii,
)

try:
    from _json import encode_basestring_ascii as c_encode_basestring_ascii
except ImportError:
    c_encode_basestring_ascii = None
try:
    from _json import encode_basestring as c_encode_basestring
except ImportError:
    c_encode_basestring = None
try:
    from _json import make_encoder as c_make_encoder
except ImportError:
    c_make_encoder = None

INFINITY = float("inf")
encode_basestring = c_encode_basestring or py_encode_basestring
encode_basestring_ascii = c_encode_basestring_ascii or py_encode_basestring_ascii


class IgnoreNanEncoder(JSONEncoder):
    def iterencode(self, o, _one_shot=False):
        """Encode the given object and yield each string
        representation as available.
        For example::
            for chunk in JSONEncoder().iterencode(bigobject):
                mysocket.write(chunk)
        """
        if self.check_circular:
            markers = {}
        else:
            markers = None
        if self.ensure_ascii:
            _encoder = encode_basestring_ascii
        else:
            _encoder = encode_basestring

        def floatstr(
            o, _repr=float.__repr__, _inf=INFINITY, _neginf=-INFINITY,
        ):
            if o != o or o == _inf or o == _neginf:
                return "null"
            else:
                return _repr(o)

        _iterencode = _make_iterencode(
            markers,
            self.default,
            _encoder,
            self.indent,
            floatstr,
            self.key_separator,
            self.item_separator,
            self.sort_keys,
            self.skipkeys,
            _one_shot,
        )
        return _iterencode(o, 0)



from tamr_unify_client.project.step import ProjectStep


class DatasetUse:
    """
    The use of a dataset in project steps. This is not a `BaseResource` because it has no API path
    and cannot be directly retrieved or modified.

    See https://docs.tamr.com/reference#retrieve-downstream-dataset-usage

    :param client: Delegate underlying API calls to this client.
    :type client: :class:`~tamr_unify_client.Client`
    :param data: The JSON body containing usage information.
    :type data: :py:class:`dict`
    """

    def __init__(self, client, data):
        self.client = client
        self._data = data

    @property
    def dataset_id(self):
        """:type: str"""
        return self._data.get("datasetId")

    @property
    def dataset_name(self):
        """:type: str"""
        return self._data.get("datasetName")

    @property
    def input_to_project_steps(self):
        """:type: list[:class:`~tamr_unify_client.project.step.ProjectStep`]"""
        steps = self._data.get("inputToProjectSteps")
        return [ProjectStep(self.client, step) for step in steps]

    @property
    def output_from_project_steps(self):
        """:type: list[:class:`~tamr_unify_client.project.step.ProjectStep`]"""
        steps = self._data.get("outputFromProjectSteps")
        return [ProjectStep(self.client, step) for step in steps]

    def dataset(self):
        """Retrieves the :class:`~tamr_unify_client.dataset.resource.Dataset` this use represents.

        :return: The dataset being used.
        :rtype: :class:`~tamr_unify_client.dataset.resource.Dataset`
        """
        dataset_id = self.dataset_id.split("/")[-1]
        return self.client.datasets.by_resource_id(dataset_id)

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"dataset_id={self.dataset_id!r})"
        )



from dataclasses import dataclass
from typing import Optional, Union

from tamr_client._types.url import URL


@dataclass(frozen=True)
class CategorizationProject:
    """A Tamr Categorization project

    See https://docs.tamr.com/reference/the-project-object

    Args:
        url
        name
        description
    """

    url: URL
    name: str
    description: Optional[str] = None


@dataclass(frozen=True)
class MasteringProject:
    """A Tamr Mastering project

    See https://docs.tamr.com/reference/the-project-object

    Args:
        url
        name
        description
    """

    url: URL
    name: str
    description: Optional[str] = None


@dataclass(frozen=True)
class SchemaMappingProject:
    """A Tamr Schema Mapping project

    See https://docs.tamr.com/reference/the-project-object

    Args:
        url
        name
        description
    """

    url: URL
    name: str
    description: Optional[str] = None


@dataclass(frozen=True)
class GoldenRecordsProject:
    """A Tamr Golden Records project

    See https://docs.tamr.com/reference/the-project-object

    Args:
        url
        name
        description
    """

    url: URL
    name: str
    description: Optional[str] = None


Project = Union[
    CategorizationProject, MasteringProject, SchemaMappingProject, GoldenRecordsProject
]



from tamr_unify_client.attribute.resource import Attribute
from tamr_unify_client.base_collection import BaseCollection


class AttributeCollection(BaseCollection):
    """Collection of :class:`~tamr_unify_client.attribute.resource.Attribute` s.

    :param client: Client for API call delegation.
    :type client: :class:`~tamr_unify_client.Client`
    :param api_path: API path used to access this collection.
        E.g. ``"datasets/1/attributes"``.
    :type api_path: str
    """

    def __init__(self, client, api_path):
        super().__init__(client, api_path)

    def by_resource_id(self, resource_id):
        """Retrieve an attribute by resource ID.

        :param resource_id: The resource ID. E.g. ``"AttributeName"``
        :type resource_id: str
        :returns: The specified attribute.
        :rtype: :class:`~tamr_unify_client.attribute.resource.Attribute`
        """
        return super().by_resource_id(self.api_path, resource_id)

    def by_relative_id(self, relative_id):
        """Retrieve an attribute by relative ID.

        :param relative_id: The resource ID. E.g. ``"datasets/1/attributes/AttributeName"``
        :type relative_id: str
        :returns: The specified attribute.
        :rtype: :class:`~tamr_unify_client.attribute.resource.Attribute`
        """
        return super().by_relative_id(Attribute, relative_id)

    def by_external_id(self, external_id):
        """Retrieve an attribute by external ID.

        Since attributes do not have external IDs, this method is not supported and will
        raise a :class:`NotImplementedError` .

        :param external_id: The external ID.
        :type external_id: str
        :returns: The specified attribute, if found.
        :rtype: :class:`~tamr_unify_client.attribute.resource.Attribute`
        :raises KeyError: If no attribute with the specified external_id is found
        :raises LookupError: If multiple attributes with the specified external_id are found
        """
        raise NotImplementedError("Attributes do not have external_id")

    def stream(self):
        """Stream attributes in this collection. Implicitly called when iterating
        over this collection.

        :returns: Stream of attributes.
        :rtype: Python generator yielding :class:`~tamr_unify_client.attribute.resource.Attribute`

        Usage:
            >>> for attribute in collection.stream(): # explicit
            >>>     do_stuff(attribute)
            >>> for attribute in collection: # implicit
            >>>     do_stuff(attribute)
        """
        data = self.client.get(self.api_path).successful().json()
        for resource_json in data:
            alias = self.api_path + "/" + resource_json["name"]
            yield Attribute.from_json(self.client, resource_json, alias)

    def by_name(self, attribute_name):
        """Lookup a specific attribute in this collection by exact-match on name.

        :param attribute_name: Name of the desired attribute.
        :type attribute_name: str
        :return: Attribute with matching name in this collection.
        :rtype: :class:`~tamr_unify_client.attribute.resource.Attribute`
        """
        return super().by_resource_id(self.api_path, attribute_name)

    def create(self, creation_spec):
        """
        Create an Attribute in this collection

        :param creation_spec: Attribute creation specification should be formatted as specified in the `Public Docs for adding an Attribute <https://docs.tamr.com/reference#add-attributes>`_.
        :type creation_spec: dict[str, str]
        :returns: The created Attribute
        :rtype: :class:`~tamr_unify_client.attribute.resource.Attribute`
        """
        data = self.client.post(self.api_path, json=creation_spec).successful().json()
        alias = self.api_path + "/" + creation_spec["name"]
        return Attribute.from_json(self.client, data, alias)

    # super.__repr__ is sufficient



from copy import deepcopy
from typing import List

from tamr_client import Backup, response
from tamr_client._types import Instance, JsonDict, Session, URL
from tamr_client.exception import TamrClientException


class InvalidOperation(TamrClientException):
    """Raised when attempting an invalid operation.
    """

    pass


class NotFound(TamrClientException):
    """Raised when referencing a backup that does not exist on the server.
    """

    pass


def _from_json(url: URL, data: JsonDict) -> Backup:
    """Make backup from JSON data (deserialize).

    Args:
        url: Backup URL
        data: Backup JSON data from Tamr server
    """
    cp = deepcopy(data)
    return Backup(
        url=url,
        path=cp["backupPath"],
        state=cp["state"],
        error_message=cp["errorMessage"],
    )


def get_all(session: Session, instance: Instance) -> List[Backup]:
    """Get all backups that have been initiated for a Tamr instance.

    Args:
        session: Tamr session
        instance: Tamr instance

    Returns:
        A list of Tamr backups

    Raises:
        backup.NotFound: If no backup found at the specified URL
    """
    url = URL(instance=instance, path="backups")
    r = session.get(str(url))
    if r.status_code == 404:
        raise NotFound(str(url))
    backups = [
        _from_json(URL(instance=instance, path=f'backups/{data["relativeId"]}'), data)
        for data in response.successful(r).json()
    ]
    return backups


def by_resource_id(session: Session, instance: Instance, resource_id: str) -> Backup:
    """Get information on a specific Tamr backup.

    Args:
        session: Tamr session
        instance: Tamr instance
        resource_id: Resource ID of the backup

    Returns:
        A Tamr backup

    Raises:
        backup.NotFound: If no backup found at the specified URL
    """
    url = URL(instance=instance, path=f"backups/{resource_id}")
    r = session.get(str(url))
    if r.status_code == 404:
        raise NotFound(str(url))
    return _from_json(url, response.successful(r).json())


def initiate(session: Session, instance: Instance) -> Backup:
    """Initiate a Tamr backup.

    Args:
        session: Tamr session
        instance: Tamr instance

    Returns:
        Initiated backup

    Raises:
        backup.InvalidOperation: If attempting an invalid operation
    """
    url = URL(instance=instance, path="backups")
    r = session.post(str(url))
    if r.status_code == 400:
        raise InvalidOperation(str(url), r.json()["message"])
    data = response.successful(r).json()
    return _from_json(
        URL(instance=instance, path=f'backups/{data["relativeId"]}'), data
    )


def cancel(session: Session, backup: Backup) -> Backup:
    """Cancel a Tamr backup.

    Args:
        session: Tamr session
        backup: A Tamr backup

    Returns:
        Canceled backup

    Raises:
        backup.NotFound: If no backup found at the specified URL
        backup.InvalidOperation: If attempting an invalid operation
    """
    cancel_url = f"{backup.url}:cancel"
    r = session.post(cancel_url)
    if r.status_code == 404:
        raise NotFound(cancel_url)
    if r.status_code == 400:
        raise InvalidOperation(cancel_url, r.json()["message"])
    return _from_json(backup.url, response.successful(r).json())


def poll(session: Session, backup: Backup) -> Backup:
    """Poll this backup for server-side updates.

    Does not update the :class:`~tamr_client.backup.Backup` object.
    Instead, returns a new :class:`~tamr_client.backup.Backup`.

    Args:
        session: Tamr session
        backup: Tamr backup to be polled

    Returns:
        A Tamr backup

    Raises:
        backup.NotFound: If no backup found at the specified URL
    """
    url = backup.url
    r = session.get(str(url))
    if r.status_code == 404:
        raise NotFound(str(url))
    return _from_json(url, response.successful(r).json())



from time import sleep, time as now

from tamr_unify_client.base_resource import BaseResource


class Operation(BaseResource):
    """A long-running operation performed by Tamr.
    Operations appear on the "Jobs" page of the Tamr UI.

    By design, client-side operations represent server-side operations *at a
    particular point in time* (namely, when the operation was fetched from the
    server). In other words: Operations *will not* pick up on server-side
    changes automatically. To get an up-to-date representation, refetch the
    operation e.g. ``op = op.poll()``.
    """

    @classmethod
    def from_json(cls, client, resource_json, api_path=None):
        return super().from_data(client, resource_json, api_path)

    @classmethod
    def from_resource_id(cls, client, resource_id):
        """Get an operation by resource ID.

        :param client: Delegate underlying API calls to this client.
        :type client: :class:`~tamr_unify_client.Client`
        :param resource_id: The ID of the operation
        :type resource_id: str
        :returns: The specified operation
        :rtype: :class:`~tamr_unify_client.operation.Operation`
        """
        url = f"operations/{resource_id}"
        response = client.get(url).successful()
        return Operation.from_response(client, response)

    @classmethod
    def from_response(cls, client, response):
        """
        Handle idiosyncrasies in constructing Operations from Tamr responses.
        When a Tamr API call would start an operation, but all results that would be
        produced by that operation are already up-to-date, Tamr returns `HTTP 204 No Content`

        To make it easy for client code to handle these API responses without checking
        the response code, this method will either construct an Operation, or a
        dummy `NoOp` operation representing the 204 Success response.

        :param client: Delegate underlying API calls to this client.
        :type client: :class:`~tamr_unify_client.Client`
        :param response: HTTP Response from the request that started the operation.
        :type response: :class:`requests.Response`
        :return: Operation
        :rtype: :class:`~tamr_unify_client.operation.Operation`
        """
        if response.status_code == 204:
            # Operation was successful, but the response contains no content.
            # Create a dummy operation to represent this.
            _never = "0000-00-00T00:00:00.000Z"
            _description = """Tamr returned HTTP 204 for this operation, indicating that all
                results that would be produced by the operation are already up-to-date."""
            resource_json = {
                "id": "-1",
                "type": "NOOP",
                "description": _description,
                "status": {
                    "state": "SUCCEEDED",
                    "startTime": _never,
                    "endTime": _never,
                    "message": "",
                },
                "created": {"username": "", "time": _never, "version": "-1"},
                "lastModified": {"username": "", "time": _never, "version": "-1"},
                "relativeId": "operations/-1",
            }
        else:
            resource_json = response.json()
        return Operation.from_json(client, resource_json)

    def apply_options(self, asynchronous=False, **options):
        """Applies operation options to this operation.

        **NOTE**: This function **should not** be called directly. Rather, options should be
        passed in through a higher-level function e.g. :func:`~tamr_unify_client.dataset.resource.Dataset.refresh` .

        Synchronous mode:
            Automatically waits for operation to resolve before returning the
            operation.

        asynchronous mode:
            Immediately return the ``'PENDING'`` operation. It is
            up to the user to coordinate this operation with their code via
            :func:`~tamr_unify_client.operation.Operation.wait` and/or
            :func:`~tamr_unify_client.operation.Operation.poll` .

        :param asynchronous: Whether or not to run in asynchronous mode. Default: ``False``.
        :type asynchronous: bool
        :param ``**options``: When running in synchronous mode, these options are
            passed to the underlying :func:`~tamr_unify_client.operation.Operation.wait` call.
        :return: Operation with options applied.
        :rtype: :class:`~tamr_unify_client.operation.Operation`
        """
        if asynchronous:
            return self
        return self.wait(**options)

    @property
    def type(self):
        """:type: str"""
        return self._data.get("type")

    @property
    def description(self):
        """:type: str"""
        return self._data.get("description")

    @property
    def status(self):
        return self._data.get("status")

    @property
    def state(self):
        """Server-side state of this operation.

        Operation state can be unresolved (i.e. ``state`` is one of: ``'PENDING'``, ``'RUNNING'``),
        or resolved (i.e. `state` is one of: ``'CANCELED'``, ``'SUCCEEDED'``, ``'FAILED'``).
        Unless opting into asynchronous mode, all exposed operations should be resolved.

        Note: you only need to manually pick up server-side changes when opting into asynchronous mode when kicking off this operation.

        Usage:
            >>> op.state # operation is currently 'PENDING'
            'PENDING'
            >>> op.wait() # continually polls until operation resolves
            >>> op.state # incorrect usage; operation object state never changes.
            'PENDING'
            >>> op = op.poll() # correct usage; use value returned by Operation.poll or Operation.wait
            >>> op.state
            'SUCCEEDED'
        """
        return (self.status or {}).get("state")

    def poll(self):
        """Poll this operation for server-side updates.

        Does not update the calling :class:`~tamr_unify_client.operation.Operation` object.
        Instead, returns a new :class:`~tamr_unify_client.operation.Operation`.

        :return: Updated representation of this operation.
        :rtype: :class:`~tamr_unify_client.operation.Operation`
        """
        op_json = self.client.get(self.api_path).successful().json()
        return Operation.from_json(self.client, op_json)

    def wait(self, poll_interval_seconds=3, timeout_seconds=None):
        """Continuously polls for this operation's server-side state.

        :param int poll_interval_seconds: Time interval (in seconds) between subsequent polls.
        :param int timeout_seconds: Time (in seconds) to wait for operation to resolve.
        :raises TimeoutError: If operation takes longer than `timeout_seconds` to resolve.
        :return: Resolved operation.
        :rtype: :class:`~tamr_unify_client.operation.Operation`
        """
        started = now()
        op = self
        while timeout_seconds is None or now() - started < timeout_seconds:
            if op.state in ["PENDING", "RUNNING"]:
                sleep(poll_interval_seconds)
            elif op.state in ["CANCELED", "SUCCEEDED", "FAILED"]:
                return op
            op = op.poll()
        raise TimeoutError(
            f"Waiting for operation took longer than {timeout_seconds} seconds."
        )

    def succeeded(self):
        """Convenience method for checking if operation was successful.

        :return: ``True`` if operation's state is ``'SUCCEEDED'``, ``False`` otherwise.
        :rtype: :py:class:`bool`
        """
        return self.state == "SUCCEEDED"

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"relative_id={self.relative_id!r}, "
            f"description={self.description!r}, "
            f"state={self.state!r})"
        )



from tamr_client.attribute import sub
from tamr_client.attribute import type
from tamr_client.attribute._attribute import (
    _from_json,
    AlreadyExists,
    by_resource_id,
    create,
    delete,
    NotFound,
    ReservedName,
    to_json,
    update,
)



from copy import deepcopy
import json
import os
from typing import Optional, TYPE_CHECKING
import warnings

from tamr_unify_client._ignore_nan_encoder import IgnoreNanEncoder
from tamr_unify_client.attribute.collection import AttributeCollection
from tamr_unify_client.base_resource import BaseResource
from tamr_unify_client.dataset.profile import DatasetProfile
from tamr_unify_client.dataset.status import DatasetStatus
from tamr_unify_client.dataset.uri import DatasetURI
from tamr_unify_client.dataset.usage import DatasetUsage
from tamr_unify_client.operation import Operation

BUILDING_DOCS = os.environ.get("TAMR_CLIENT_DOCS") == "1"
if TYPE_CHECKING or BUILDING_DOCS:
    import pandas as pd


class Dataset(BaseResource):
    """A Tamr dataset."""

    @classmethod
    def from_json(cls, client, resource_json, api_path=None):
        return super().from_data(client, resource_json, api_path)

    @property
    def name(self):
        """:type: str"""
        return self._data.get("name")

    @property
    def external_id(self):
        """:type: str"""
        return self._data.get("externalId")

    @property
    def description(self):
        """:type: str"""
        return self._data.get("description")

    @property
    def version(self):
        """:type: str"""
        return self._data.get("version")

    @property
    def tags(self):
        """:type: list[str]"""
        return self._data.get("tags")[:]

    @property
    def key_attribute_names(self):
        """:type: list[str]"""
        return self._data.get("keyAttributeNames")[:]

    @property
    def attributes(self):
        """Attributes of this dataset.

        :return: Attributes of this dataset.
        :rtype: :class:`~tamr_unify_client.attribute.collection.AttributeCollection`
        """
        alias = self.api_path + "/attributes"
        return AttributeCollection(self.client, alias)

    def _update_records(self, updates, *, ignore_nan=False):
        """Send a batch of record creations/updates/deletions to this dataset.
        You probably want to use :func:`~tamr_unify_client.dataset.resource.Dataset.upsert_records`
        or :func:`~tamr_unify_client.dataset.resource.Dataset.delete_records` instead.

        :param records: Each record should be formatted as specified in the `Public Docs for Dataset updates <https://docs.tamr.com/reference#modify-a-datasets-records>`_.
        :type records: iterable[dict]
        :param ignore_nan: Whether to treat `NaN` values as null. Unconverted `NaN`s will raise an error if found. Deprecated.
        :type ignore_nan: bool
        :returns: JSON response body from server.
        :rtype: :py:class:`dict`
        """
        if ignore_nan:
            warnings.warn(
                "'ignore_nan' is deprecated. Users are expected to provide valid JSON representations instead",
                DeprecationWarning,
            )
        encoder = IgnoreNanEncoder if ignore_nan else None
        stringified_updates = (
            json.dumps(update, cls=encoder, allow_nan=False).encode("utf-8")
            for update in updates
        )

        return (
            self.client.post(
                self.api_path + ":updateRecords",
                headers={"Content-Encoding": "utf-8"},
                data=stringified_updates,
            )
            .successful()
            .json()
        )

    def upsert_from_dataframe(
        self,
        df: "pd.DataFrame",
        *,
        primary_key_name: str,
        ignore_nan: Optional[bool] = None,
    ) -> dict:
        """Upserts a record for each row of `df` with attributes for each column in `df`.

        Args:
            df: The data to upsert records from.
            primary_key_name: The name of the primary key of the dataset.  Must be a column of `df`.
            ignore_nan: Legacy parameter that does nothing. Deprecated.

        Returns:
            JSON response body from the server.

        Raises:
            KeyError: If `primary_key_name` is not a column in `df`.

        """
        if ignore_nan is not None:
            warnings.warn(
                "'ignore_nan' is deprecated. DataFrame `NaN`s are always ignored in upsert",
                DeprecationWarning,
            )
        if primary_key_name not in df.columns:
            raise KeyError(f"{primary_key_name} is not an attribute of the data")

        # serialize records via to_json to handle `np.nan` values
        serialized_records = ((pk, row.to_json()) for pk, row in df.iterrows())
        records = (
            {primary_key_name: pk, **json.loads(row)} for pk, row in serialized_records
        )
        return self.upsert_records(records, primary_key_name)

    def upsert_records(self, records, primary_key_name, *, ignore_nan=False):
        """Creates or updates the specified records.

        :param records: The records to update, as dictionaries.
        :type records: iterable[dict]
        :param primary_key_name: The name of the primary key for these records, which must be a key in each record dictionary.
        :type primary_key_name: str
        :param ignore_nan: Whether to convert `NaN` values to `null` when upserting records.  If `False` and `NaN` is found this function will fail. Deprecated.
        :type ignore_nan: bool
        :return: JSON response body from the server.
        :rtype: dict
        """
        if ignore_nan:
            warnings.warn(
                "'ignore_nan' is deprecated. Users are expected to provide valid JSON representations instead",
                DeprecationWarning,
            )
        updates = (
            {"action": "CREATE", "recordId": record[primary_key_name], "record": record}
            for record in records
        )
        return self._update_records(updates, ignore_nan=ignore_nan)

    def delete_records(self, records, primary_key_name):
        """Deletes the specified records.

        :param records: The records to delete, as dictionaries.
        :type records: iterable[dict]
        :param primary_key_name: The name of the primary key for these records, which must be a key in each record dictionary.
        :type primary_key_name: str
        :return: JSON response body from the server.
        :rtype: dict
        """
        ids = (record[primary_key_name] for record in records)
        return self.delete_records_by_id(ids)

    def delete_records_by_id(self, record_ids):
        """Deletes the specified records.

        :param record_ids: The IDs of the records to delete.
        :type record_ids: iterable
        :return: JSON response body from the server.
        :rtype: dict
        """
        updates = ({"action": "DELETE", "recordId": rid} for rid in record_ids)
        return self._update_records(updates)

    def delete_all_records(self):
        """Removes all records from the dataset.

        :return: HTTP response from the server
        :rtype: :class:`requests.Response`
        """
        path = self.api_path + "/records"
        response = self.client.delete(path).successful()
        return response

    def refresh(self, **options):
        """Brings dataset up-to-date if needed, taking whatever actions are required.

        :param ``**options``: Options passed to underlying :class:`~tamr_unify_client.operation.Operation` .
            See :func:`~tamr_unify_client.operation.Operation.apply_options` .
        :returns: The refresh operation.
        :rtype: :class:`~tamr_unify_client.operation.Operation`
        """
        response = self.client.post(self.api_path + ":refresh").successful()
        op = Operation.from_response(self.client, response)
        return op.apply_options(**options)

    def profile(self):
        """Returns profile information for a dataset.

        If profile information has not been generated, call create_profile() first.
        If the returned profile information is out-of-date, you can call refresh() on the returned
        object to bring it up-to-date.

        :return: Dataset Profile information.
        :rtype: :class:`~tamr_unify_client.dataset.profile.DatasetProfile`
        """
        profile_json = self.client.get(self.api_path + "/profile").successful().json()
        return DatasetProfile.from_json(
            self.client, profile_json, api_path=self.api_path + "/profile"
        )

    def create_profile(self, **options):
        """Create a profile for this dataset.

        If a profile already exists, the existing profile will be brought
        up to date.

        :param ``**options``: Options passed to underlying :class:`~tamr_unify_client.operation.Operation` .
            See :func:`~tamr_unify_client.operation.Operation.apply_options` .
        :return: The operation to create the profile.
        :rtype: :class:`~tamr_unify_client.operation.Operation`
        """
        response = self.client.post(self.api_path + "/profile:refresh").successful()
        op = Operation.from_response(self.client, response)
        return op.apply_options(**options)

    def records(self):
        """Stream this dataset's records as Python dictionaries.

        :return: Stream of records.
        :rtype: Python generator yielding :py:class:`dict`
        """
        with self.client.get(self.api_path + "/records", stream=True) as response:
            for line in response.iter_lines():
                yield json.loads(line)

    def status(self):
        """Retrieve this dataset's streamability status.

        :return: Dataset streamability status.
        :rtype: :class:`~tamr_unify_client.dataset.status.DatasetStatus`
        """
        status_json = self.client.get(self.api_path + "/status").successful().json()
        return DatasetStatus.from_json(
            self.client, status_json, api_path=self.api_path + "/status"
        )

    def usage(self):
        """Retrieve this dataset's usage by recipes and downstream datasets.

        :return: The dataset's usage.
        :rtype: :class:`~tamr_unify_client.dataset.usage.DatasetUsage`
        """
        alias = self.api_path + "/usage"
        usage = self.client.get(alias).successful().json()
        return DatasetUsage.from_json(self.client, usage, alias)

    def from_geo_features(self, features, geo_attr=None):
        """Upsert this dataset from a geospatial FeatureCollection or iterable of Features.

        `features` can be:

        - An object that implements ``__geo_interface__`` as a FeatureCollection
          (see https://gist.github.com/sgillies/2217756)
        - An iterable of features, where each element is a feature dictionary or an object
          that implements the ``__geo_interface__`` as a Feature
        - A map where the "features" key contains an iterable of features

        See: geopandas.GeoDataFrame.from_features()

        If geo_attr is provided, then the named Tamr attribute will be used for the geometry.
        If geo_attr is not provided, then the first attribute on the dataset with geometry type
        will be used for the geometry.

        :param features: geospatial features
        :param geo_attr: (optional) name of the Tamr attribute to use for the feature's geometry
        :type geo_attr: str
        :returns: JSON response body from server.
        :rtype: :py:class:`dict`
        """
        if hasattr(features, "__geo_interface__"):
            features = features.__geo_interface__
        if hasattr(features, "get") and features.get("type") == "FeatureCollection":
            features = features["features"]

        key_attrs = self.key_attribute_names
        if len(key_attrs) == 1:
            record_id = "recordId"
        else:
            record_id = "compositeRecordId"

        if geo_attr is None:
            geo_attr = self._geo_attr

        return self._update_records(
            self._features_to_updates(features, record_id, key_attrs, geo_attr)
        )

    def upstream_datasets(self):
        """The Dataset's upstream datasets.

        API returns the URIs of the upstream datasets,
        resulting in a list of DatasetURIs, not actual Datasets.

        :return: A list of the Dataset's upstream datasets.
        :rtype: list[:class:`~tamr_unify_client.dataset.uri.DatasetURI`]
        """
        alias = self.api_path + "/upstreamDatasets"
        resources = self.client.get(alias).successful().json()

        return [DatasetURI(self.client, uri) for uri in resources]

    def spec(self):
        """Returns this dataset's spec.

        :return: The spec of this dataset.
        :rtype: :class:`~tamr_unify_client.dataset.resource.DatasetSpec`
        """
        return DatasetSpec.of(self)

    def delete(self, cascade=False):
        """Deletes this dataset, optionally deleting all derived datasets as well.

        :param cascade: Whether to delete all datasets derived from this one. Optional, default is `False`.
            Do not use this option unless you are certain you need it as it can have unindended consequences.
        :type cascade: bool
        :return: HTTP response from the server
        :rtype: :class:`requests.Response`
        """
        params = {"cascade": cascade}
        response = self.client.delete(self.api_path, params=params).successful()
        return response

    @property
    def __geo_interface__(self):
        """Retrieve a representation of this dataset that conforms to the Python Geo Interface.

        Note that this materializes all features; for a streaming interface to features,
        see :method:`~tamr_unify_client.dataset.Dataset.__geo_features__()`

        See https://gist.github.com/sgillies/2217756

        :return: dict[str, object]
        """
        return {
            "type": "FeatureCollection",
            "features": [feature for feature in self.itergeofeatures()],
        }

    def itergeofeatures(self, geo_attr=None):
        """Returns an iterator that yields feature dictionaries that comply with __geo_interface__

        See https://gist.github.com/sgillies/2217756

        :param geo_attr: (optional) name of the Tamr attribute to use for the feature's geometry
        :type geo_attr: str
        :return: stream of features
        :rtype: Python generator yielding :py:class:`dict[str, object]`
        """
        key_attrs = self.key_attribute_names
        if len(key_attrs) == 1:

            def key_value(rec):
                return rec[key_attrs[0]]

        else:

            def key_value(rec):
                return [rec[attr] for attr in key_attrs]

        if geo_attr is None:
            geo_attr = self._geo_attr

        for record in self.records():
            yield self._record_to_feature(record, key_value, key_attrs, geo_attr)

    @property
    def _geo_attr(self):
        """The name of the attribute that contains geometry

        :return: the name of the attribute that contains geometry
        :rtype: str
        """
        # Duck-typing: find all the attributes that look like geometry
        geo_attrs = [
            attr.name
            for attr in self.attributes
            if "RECORD" == attr.type.base_type
            and self._geo_attr_names().intersection(
                {sub_attr.name for sub_attr in attr.type.attributes}
            )
        ]
        # We select the first such attribute as the geometry
        if geo_attrs:
            geo_attr = geo_attrs[0]
        else:
            geo_attr = None
        return geo_attr

    @staticmethod
    def _record_to_feature(record, key_value, key_attrs, geo_attr):
        """Convert a Tamr record to a Python Geo Interface Feature

        :param record: Tamr record
        :param key_value: Function to extract the value of the primary key from the record
        :param key_attrs: Set of attributes that comprise the primary key for the record
        :param geo_attr: The singular attribute to use as the geometry
        :return: map from str to object
        """
        feature = {"type": "Feature", "id": key_value(record)}
        reserved = {"bbox", geo_attr}.union(key_attrs)
        if geo_attr and geo_attr in record:
            src_geo = record[geo_attr]
            if src_geo:
                for unify_attr in Dataset._geo_attr_names():
                    if unify_attr in src_geo and src_geo[unify_attr]:
                        feature["geometry"] = {
                            # Convert e.g. multiLineString -> MultiLineString
                            "type": unify_attr[0].upper() + unify_attr[1:],
                            "coordinates": src_geo[unify_attr],
                        }
                        break
                    else:
                        feature["geometry"] = None
            else:
                feature["geometry"] = None
        if "bbox" in record:
            feature["bbox"] = record["bbox"]
        non_reserved = set(record.keys()).difference(reserved)
        if non_reserved:
            feature["properties"] = {attr: record[attr] for attr in non_reserved}
        return feature

    @staticmethod
    def _feature_to_record(feature, key_attrs, geo_attr):
        """Convert a Python Geo Interface Feature to a Tamr record

        feature can be a dict representing a Geospatial Feature, or a Feature object
        that implements the __geo_interface__ property.

        :param feature: Python Geo Interface Feature
        :param key_attrs: Sequence of attributes that comprise the primary key for the record
        :param geo_attr: The singluar attribute on the record to use for the geometry
        :return: dict
        """

        if hasattr(feature, "__geo_interface__"):
            feature = feature.__geo_interface__

        record = {}

        props = feature.get("properties")
        if props:
            for prop in props:
                record[prop] = props[prop]

        geometry = feature.get("geometry")
        if geometry:
            geo_type = geometry["type"]
            # Convert e.g. "MultiLineString" -> "multiLineString"
            geo_type = geo_type[0].lower() + geo_type[1:]
            record[geo_attr] = {geo_type: geometry["coordinates"]}

        bbox = feature.get("bbox")
        if bbox:
            record["bbox"] = bbox
        if "id" not in feature or feature["id"] is None:
            raise ValueError("id must have a non-null value")
        if key_attrs[1:]:
            key_values = feature["id"]

            for i, attr in enumerate(key_attrs):
                record[attr] = key_values[i]
        else:
            record[key_attrs[0]] = feature["id"]
        return record

    @staticmethod
    def _features_to_updates(features, id_attr, key_attrs, geo_attr):
        for feature in features:
            yield {
                "action": "CREATE",
                id_attr: feature["id"],
                "record": Dataset._feature_to_record(feature, key_attrs, geo_attr),
            }

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"relative_id={self.relative_id!r}, "
            f"name={self.name!r}, "
            f"version={self.version!r})"
        )

    @staticmethod
    def _geo_attr_names():
        return {
            "point",
            "multiPoint",
            "lineString",
            "multiLineString",
            "polygon",
            "multiPolygon",
        }


class DatasetSpec:
    """A representation of the server view of a dataset."""

    def __init__(self, client, data, api_path):
        self.client = client
        self._data = data
        self.api_path = api_path

    @staticmethod
    def of(resource):
        """Creates a dataset spec from a dataset.

        :param resource: The existing dataset.
        :type resource: :class:`~tamr_unify_client.dataset.resource.Dataset`
        :return: The corresponding dataset spec.
        :rtype: :class:`~tamr_unify_client.dataset.resource.DatasetSpec`
        """
        return DatasetSpec(resource.client, deepcopy(resource._data), resource.api_path)

    @staticmethod
    def new():
        """Creates a blank spec that could be used to construct a new dataset.

        :return: The empty spec.
        :rtype: :class:`~tamr_unify_client.dataset.resource.DatasetSpec`
        """
        return DatasetSpec(None, {}, None)

    def from_data(self, data):
        """Creates a spec with the same client and API path as this one, but new data.

        :param data: The data for the new spec.
        :type data: dict
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.dataset.resource.DatasetSpec`
        """
        return DatasetSpec(self.client, data, self.api_path)

    def to_dict(self):
        """Returns a version of this spec that conforms to the API representation.

        :returns: The spec's dict.
        :rtype: dict
        """
        return deepcopy(self._data)

    def with_name(self, new_name):
        """Creates a new spec with the same properties, updating name.

        :param new_name: The new name.
        :type new_name: str
        :return: A new spec.
        :rtype: :class:`~tamr_unify_client.dataset.resource.DatasetSpec`
        """
        return self.from_data({**self._data, "name": new_name})

    def with_external_id(self, new_external_id):
        """Creates a new spec with the same properties, updating external ID.

        :param new_external_id: The new external ID.
        :type new_external_id: str
        :return: A new spec.
        :rtype: :class:`~tamr_unify_client.dataset.resource.DatasetSpec`
        """
        return self.from_data({**self._data, "externalId": new_external_id})

    def with_description(self, new_description):
        """Creates a new spec with the same properties, updating description.

        :param new_description: The new description.
        :type new_description: str
        :return: A new spec.
        :rtype: :class:`~tamr_unify_client.dataset.resource.DatasetSpec`
        """
        return self.from_data({**self._data, "description": new_description})

    def with_key_attribute_names(self, new_key_attribute_names):
        """Creates a new spec with the same properties, updating key attribute names.

       :param new_key_attribute_names: The new key attribute names.
       :type new_key_attribute_names: list[str]
       :return: A new spec.
       :rtype: :class:`~tamr_unify_client.dataset.resource.DatasetSpec`
       """
        return self.from_data(
            {**self._data, "keyAttributeNames": new_key_attribute_names}
        )

    def with_tags(self, new_tags):
        """Creates a new spec with the same properties, updating tags.

        :param new_tags: The new tags.
        :type new_tags: list[str]
        :return: A new spec.
        :rtype: :class:`~tamr_unify_client.dataset.resource.DatasetSpec`
        """
        return self.from_data({**self._data, "tags": new_tags})

    def put(self):
        """Updates the dataset on the server.

        :return: The modified dataset.
        :rtype: :class:`~tamr_unify_client.dataset.resource.Dataset`
        """
        new_data = self.client.put(self.api_path, json=self._data).successful().json()
        return Dataset.from_json(self.client, new_data, self.api_path)

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"dict={self._data})"
        )



"""
Tamr - Categorization
See https://docs.tamr.com/docs/overall-workflow-classification
"""
from tamr_client.categorization import project
from tamr_client.categorization._categorization import (
    _apply_feedback_async,
    _update_results_async,
    apply_feedback,
    manual_labels,
    update_results,
    update_unified_dataset,
)



"""
Tamr - Mastering
See https://docs.tamr.com/docs/overall-workflow-mastering

The terminology used here is consistent with Tamr UI terminology

Asynchronous versions of each function can be found with the suffix `_async` and may be of
interest to power users
"""
from tamr_client import operation
from tamr_client._types import MasteringProject, Operation, Session
from tamr_client.dataset import unified


def update_unified_dataset(session: Session, project: MasteringProject) -> Operation:
    """Apply changes to the unified dataset and wait for the operation to complete

    Args:
        project: Tamr Mastering project
    """
    unified_dataset = unified.from_project(session, project)
    op = unified._apply_changes_async(session, unified_dataset)
    return operation.wait(session, op)


def estimate_pairs(session: Session, project: MasteringProject) -> Operation:
    """Update the estimated pair counts and wait for the operation to complete

    Args:
        project: Tamr Mastering project
    """
    op = _estimate_pairs_async(session, project)
    return operation.wait(session, op)


def generate_pairs(session: Session, project: MasteringProject) -> Operation:
    """Generate pairs according to the binning model and wait for the operation
    to complete

    Args:
        project: Tamr Mastering project
    """
    op = _generate_pairs_async(session, project)
    return operation.wait(session, op)


def apply_feedback(session: Session, project: MasteringProject) -> Operation:
    """Train the pair-matching model according to verified labels and wait for the
    operation to complete

    Args:
        project: Tamr Mastering project
    """
    op = _apply_feedback_async(session, project)
    return operation.wait(session, op)


def update_pair_results(session: Session, project: MasteringProject) -> Operation:
    """Update record pair predictions according to the latest pair-matching model and
    wait for the operation to complete

    Args:
        project: Tamr Mastering project
    """
    op = _update_pair_results_async(session, project)
    return operation.wait(session, op)


def update_high_impact_pairs(session: Session, project: MasteringProject) -> Operation:
    """Produce new high-impact pairs according to the latest pair-matching model and
    wait for the operation to complete

    Args:
        project: Tamr Mastering project
    """
    op = _update_high_impact_pairs_async(session, project)
    return operation.wait(session, op)


def update_cluster_results(session: Session, project: MasteringProject) -> Operation:
    """Generate clusters based on the latest pair-matching model and wait for the
    operation to complete

    Args:
        project: Tamr Mastering project
    """
    op = _update_cluster_results_async(session, project)
    return operation.wait(session, op)


def publish_clusters(session: Session, project: MasteringProject) -> Operation:
    """Publish current record clusters and wait for the operation to complete

    Args:
        project: Tamr Mastering project
    """
    op = _publish_clusters_async(session, project)
    return operation.wait(session, op)


def _estimate_pairs_async(session: Session, project: MasteringProject) -> Operation:
    r = session.post(str(project.url) + "/estimatedPairCounts:refresh")
    return operation._from_response(project.url.instance, r)


def _generate_pairs_async(session: Session, project: MasteringProject) -> Operation:
    r = session.post(str(project.url) + "/recordPairs:refresh")
    return operation._from_response(project.url.instance, r)


def _apply_feedback_async(session: Session, project: MasteringProject) -> Operation:
    r = session.post(str(project.url) + "/recordPairsWithPredictions/model:refresh")
    return operation._from_response(project.url.instance, r)


def _update_pair_results_async(
    session: Session, project: MasteringProject
) -> Operation:
    r = session.post(str(project.url) + "/recordPairsWithPredictions:refresh")
    return operation._from_response(project.url.instance, r)


def _update_high_impact_pairs_async(
    session: Session, project: MasteringProject
) -> Operation:
    r = session.post(str(project.url) + "/highImpactPairs:refresh")
    return operation._from_response(project.url.instance, r)


def _update_cluster_results_async(
    session: Session, project: MasteringProject
) -> Operation:
    r = session.post(str(project.url) + "/recordClusters:refresh")
    return operation._from_response(project.url.instance, r)


def _publish_clusters_async(session: Session, project: MasteringProject) -> Operation:
    r = session.post(str(project.url) + "/publishedClustersWithData:refresh")
    return operation._from_response(project.url.instance, r)



"""
See https://docs.tamr.com/reference/dataset-models
"""
from copy import deepcopy

from tamr_client import operation, response
from tamr_client._types import (
    JsonDict,
    Operation,
    Project,
    Session,
    UnifiedDataset,
    URL,
)
from tamr_client.exception import TamrClientException


class NotFound(TamrClientException):
    """Raised when referencing (e.g. updating or deleting) a unified dataset
    that does not exist on the server.
    """

    pass


def from_project(session: Session, project: Project) -> UnifiedDataset:
    """Get unified dataset of a project

    Fetches the unified dataset of a given project from Tamr server

    Args:
        project: Tamr project of this Unified Dataset

    Raises:
        unified.NotFound: If no unified dataset could be found at the specified URL.
            Corresponds to a 404 HTTP error.
        requests.HTTPError: If any other HTTP error is encountered.
    """
    url = URL(instance=project.url.instance, path=f"{project.url.path}/unifiedDataset")
    return _by_url(session, url)


def _by_url(session: Session, url: URL) -> UnifiedDataset:
    """Get dataset by URL

    Fetches dataset from Tamr server

    Args:
        url: Dataset URL

    Raises:
        unified.NotFound: If no dataset could be found at the specified URL.
            Corresponds to a 404 HTTP error.
        requests.HTTPError: If any other HTTP error is encountered.
    """
    r = session.get(str(url))
    if r.status_code == 404:
        raise NotFound(str(url))
    data = response.successful(r).json()
    return _from_json(url, data)


def _from_json(url: URL, data: JsonDict) -> UnifiedDataset:
    """Make unified dataset from JSON data (deserialize)

    Args:
        url: Unified Dataset URL
        data: Unified Dataset JSON data from Tamr server
    """
    cp = deepcopy(data)
    return UnifiedDataset(
        url,
        name=cp["name"],
        description=cp.get("description"),
        key_attribute_names=tuple(cp["keyAttributeNames"]),
    )


def apply_changes(session: Session, unified_dataset: UnifiedDataset) -> Operation:
    """Applies changes to the unified dataset and waits for the operation to complete

    Args:
        unified_dataset: The Unified Dataset which will be committed
    """
    op = _apply_changes_async(session, unified_dataset)
    return operation.wait(session, op)


def _apply_changes_async(
    session: Session, unified_dataset: UnifiedDataset
) -> Operation:
    """Applies changes to the unified dataset

    Args:
        unified_dataset: The Unified Dataset which will be committed
    """
    r = session.post(str(unified_dataset.url) + ":refresh")
    return operation._from_response(unified_dataset.url.instance, r)



from tamr_client._types import (
    GoldenRecordsProject,
    JsonDict,
    URL,
)


def _from_json(url: URL, data: JsonDict) -> GoldenRecordsProject:
    """Make golden records project from JSON data (deserialize)

    Args:
        url: Project URL
        data: Project JSON data from Tamr server
    """
    return GoldenRecordsProject(
        url, name=data["name"], description=data.get("description")
    )



"""
Tamr - Categorization
See https://docs.tamr.com/docs/overall-workflow-classification

The terminology used here is consistent with Tamr UI terminology

Asynchronous versions of each function can be found with the suffix `_async` and may be of
interest to power users
"""
from tamr_client import operation
from tamr_client._types import CategorizationProject, Dataset, Operation, Session
from tamr_client.dataset import _dataset, unified


def manual_labels(session: Session, project: CategorizationProject) -> Dataset:
    """Get manual labels from a Categorization project.

    Args:
        project: Tamr project containing labels

    Returns:
        Dataset containing manual labels

    Raises:
        dataset.NotFound: If no dataset could be found at the specified URL
        dataset.Ambiguous: If multiple targets match dataset name
    """
    unified_dataset = unified.from_project(session=session, project=project)
    labels_dataset_name = unified_dataset.name + "_manual_categorizations"
    return _dataset.by_name(
        session=session, instance=project.url.instance, name=labels_dataset_name
    )


def update_unified_dataset(
    session: Session, project: CategorizationProject
) -> Operation:
    """Apply changes to the unified dataset and wait for the operation to complete

    Args:
        project: Tamr Categorization project
    """
    unified_dataset = unified.from_project(session, project)
    op = unified._apply_changes_async(session, unified_dataset)
    return operation.wait(session, op)


def apply_feedback(session: Session, project: CategorizationProject) -> Operation:
    """Train the categorization model according to verified labels and wait for the
    operation to complete

    Args:
        project: Tamr Categorization project
    """
    op = _apply_feedback_async(session, project)
    return operation.wait(session, op)


def update_results(session: Session, project: CategorizationProject) -> Operation:
    """Generate classifications based on the latest categorization model and wait for the
    operation to complete

    Args:
        project: Tamr Categorization project
    """
    op = _update_results_async(session, project)
    return operation.wait(session, op)


def _apply_feedback_async(
    session: Session, project: CategorizationProject
) -> Operation:
    r = session.post(str(project.url) + "/categorizations/model:refresh")
    return operation._from_response(project.url.instance, r)


def _update_results_async(
    session: Session, project: CategorizationProject
) -> Operation:
    r = session.post(str(project.url) + "/categorizations:refresh")
    return operation._from_response(project.url.instance, r)



class Metric:
    """ A metric for a published cluster.

    This is not a `BaseResource` because it does not have its own API endpoint.

    :param data: The JSON entity representing this cluster.
    """

    def __init__(self, data):
        self._data = data

    @property
    def name(self):
        """:type: str"""
        return self._data.get("metricName")

    @property
    def value(self):
        """:type: str"""
        return self._data.get("metricValue")

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"name={self.name!r}, "
            f"value={self.value!r})"
        )



"""This module and attribute_type depend on each other.

"""
from copy import deepcopy

from tamr_client._types import JsonDict, SubAttribute
from tamr_client.attribute import type as attribute_type


def from_json(data: JsonDict) -> SubAttribute:
    """Make a SubAttribute from JSON data (deserialize)

    Args:
        data: JSON data received from Tamr server.
    """

    cp = deepcopy(data)
    d = {}
    d["name"] = cp["name"]
    d["is_nullable"] = cp["isNullable"]
    d["type"] = attribute_type.from_json(cp["type"])
    return SubAttribute(**d)


def to_json(subattr: SubAttribute) -> JsonDict:
    """Serialize subattribute into JSON

    Args:
        subattr: SubAttribute to serialize
    """

    d = {
        "name": subattr.name,
        "type": attribute_type.to_json(subattr.type),
        "isNullable": subattr.is_nullable,
    }
    return d



class DatasetURI:
    """
    Indentifier of a dataset.

    :param client: Queried dataset's client.
    :type client: :class:`~tamr_unify_client.client.Client`
    :param uri: Queried dataset's dataset ID.
    :type uri: :py:class:`str`
    """

    def __init__(self, client, uri):
        self.client = client
        self._uri = uri

    @property
    def resource_id(self):
        """:type: str"""
        return self._uri.split("/")[-1]

    @property
    def relative_id(self):
        """:type: str"""
        return "datasets/" + self.resource_id

    @property
    def uri(self):
        """:type: str"""
        return self._uri

    def dataset(self):
        """Fetch the dataset that this identifier points to.

        :return: A Tamr dataset.
        :rtype: :class: `~tamr_unify_client.dataset.resource.Dataset`
        """
        return self.client.datasets.by_resource_id(self.resource_id)

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"'{self.uri})'"
        )



"""
Tamr - Golden Records
See https://docs.tamr.com/docs/overview-golden-records

The terminology used here is consistent with Tamr UI terminology

Asynchronous versions of each function can be found with the suffix `_async` and may be of
interest to power users
"""
from tamr_client import operation
from tamr_client._types import GoldenRecordsProject, Operation, Session


def update(session: Session, project: GoldenRecordsProject) -> Operation:
    """Update the draft golden records and wait for the operation to complete

    Args:
        project: Tamr Golden Records project
    """
    op = _update_async(session, project)
    return operation.wait(session, op)


def publish(session: Session, project: GoldenRecordsProject) -> Operation:
    """Publish the golden records and wait for the operation to complete

    Args:
        project: Tamr Golden Records project
    """
    op = _publish_async(session, project)
    return operation.wait(session, op)


def _update_async(session: Session, project: GoldenRecordsProject) -> Operation:
    r = session.post(str(project.url) + "/goldenRecords:refresh")
    return operation._from_response(project.url.instance, r)


def _publish_async(session: Session, project: GoldenRecordsProject) -> Operation:
    r = session.post(
        str(project.url) + "/publishedGoldenRecords:refresh",
        params={"validate": "true", "version": "CURRENT"},
    )
    return operation._from_response(project.url.instance, r)



"""
Tamr - Mastering
See https://docs.tamr.com/docs/overall-workflow-mastering
"""
from tamr_client.mastering import project
from tamr_client.mastering._mastering import (
    _apply_feedback_async,
    _estimate_pairs_async,
    _generate_pairs_async,
    _publish_clusters_async,
    _update_cluster_results_async,
    _update_high_impact_pairs_async,
    _update_pair_results_async,
    apply_feedback,
    estimate_pairs,
    generate_pairs,
    publish_clusters,
    update_cluster_results,
    update_high_impact_pairs,
    update_pair_results,
    update_unified_dataset,
)



from copy import deepcopy


class AttributeType:
    """
    The type of an :class:`~tamr_unify_client.attribute.resource.Attribute` or :class:`~tamr_unify_client.attribute.subattribute.SubAttribute`.

    See https://docs.tamr.com/reference#attribute-types

    :param data: JSON data representing this type
    :type data: :py:class:`dict`
    """

    def __init__(self, data):
        self._data = data

    @property
    def base_type(self):
        """:type: str"""
        return self._data.get("baseType")

    @property
    def inner_type(self):
        """:type: :class:`~tamr_unify_client.attribute.type.AttributeType`"""
        if "innerType" in self._data:
            return AttributeType(self._data.get("innerType"))
        else:
            return None

    @property
    def attributes(self):
        """:type: list[:class:`~tamr_unify_client.attribute.subattribute.SubAttribute`]"""
        from tamr_unify_client.attribute.subattribute import SubAttribute

        collection_json = self._data.get("attributes")
        return [SubAttribute.from_json(attr) for attr in collection_json]

    def spec(self):
        """Returns a spec representation of this attribute type.

        :return: The attribute type spec.
        :rtype: :class:`~tamr_unify_client.attribute.type.AttributeTypeSpec`
        """
        return AttributeTypeSpec.of(self)

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"base_type={self.base_type!r})"
        )


class AttributeTypeSpec:
    def __init__(self, data):
        self._data = data

    @staticmethod
    def of(resource):
        """Creates an attribute type spec from an attribute type.

        :param resource: The existing attribute type.
        :type resource: :class:`~tamr_unify_client.attribute.type.AttributeType`
        :return: The corresponding attribute type spec.
        :rtype: :class:`~tamr_unify_client.attribute.type.AttributeTypeSpec`
        """
        return AttributeTypeSpec(deepcopy(resource._data))

    @staticmethod
    def new():
        """Creates a blank spec that could be used to construct a new attribute type.

        :return: The empty spec.
        :rtype: :class:`~tamr_unify_client.attribute.type.AttributeTypeSpec`
        """
        return AttributeTypeSpec({})

    def to_dict(self):
        """Returns a version of this spec that conforms to the API representation.

        :returns: The spec's dict.
        :rtype: dict
        """
        return deepcopy(self._data)

    def with_base_type(self, new_base_type):
        """Creates a new spec with the same properties, updating the base type.

        :param new_base_type: The new base type.
        :type new_base_type: str
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.attribute.type.AttributeTypeSpec`
        """
        return AttributeTypeSpec({**self._data, "baseType": new_base_type})

    def with_inner_type(self, new_inner_type):
        """Creates a new spec with the same properties, updating the inner type.

        :param new_inner_type: The spec of the new inner type.
        :type new_inner_type: :class:`~tamr_unify_client.attribute.type.AttributeTypeSpec`
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.attribute.type.AttributeTypeSpec`
        """
        inner_spec = new_inner_type.to_dict()
        return AttributeTypeSpec({**self._data, "innerType": inner_spec})

    def with_attributes(self, new_attributes):
        """Creates a new spec with the same properties, updating attributes.

        :param new_attributes: The specs of the new attributes.
        :type new_attributes: list[:class:`~tamr_unify_client.attribute.resource.AttributeSpec`]
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.attribute.type.AttributeTypeSpec`
        """
        attr_specs = [attr.to_dict() for attr in new_attributes]
        return AttributeTypeSpec({**self._data, "attributes": attr_specs})

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"dict={self._data!r})"
        )



from tamr_unify_client.project.attribute_mapping.resource import AttributeMapping


class AttributeMappingCollection:
    """Collection of :class:`~tamr_unify_client.project.attribute_mapping.resource.AttributeMapping`

    :param client: Client for API call delegation.
    :type client: :class:`~tamr_unify_client.Client`
    :param api_path: API path used to access this collection.
    :type api_path: str
    """

    def __init__(self, client, api_path):
        self.client = client
        self.api_path = api_path

    def stream(self):
        """Stream attribute mappings in this collection. Implicitly called when iterating
        over this collection.

        :returns: Stream of attribute mappings.
        :rtype: Python generator yielding :class:`~tamr_unify_client.project.attribute_mapping.resource.AttributeMapping`
        """
        all_maps = self.client.get(self.api_path).successful().json()
        for mapping in all_maps:
            yield AttributeMapping(self.client, mapping)

    def by_resource_id(self, resource_id):
        """Retrieve an item in this collection by resource ID.

        :param resource_id: The resource ID.
        :type resource_id: str
        :returns: The specified attribute mapping.
        :rtype: :class:`~tamr_unify_client.project.attribute_mapping.resource.AttributeMapping`
        """
        maps = self.stream()
        for mapping in maps:
            split_id = mapping.resource_id
            if resource_id == split_id:
                return mapping
        raise LookupError("cannot locate mapping from resource ID")

    def by_relative_id(self, relative_id):
        """Retrieve an item in this collection by relative ID.

       :param relative_id: The relative ID.
       :type relative_id: str
       :returns: The specified attribute mapping.
       :rtype: :class:`~tamr_unify_client.project.attribute_mapping.resource.AttributeMapping`
       """
        resource_id = relative_id.split("attributeMappings/")[1]
        return self.by_resource_id(resource_id)

    def create(self, creation_spec):
        """Create an Attribute mapping in this collection

        :param creation_spec: Attribute mapping creation specification should be formatted as specified in the
            `Public Docs for adding an AttributeMapping <https://docs.tamr.com/reference#create-an-attribute-mapping>`_.
        :type creation_spec: dict[str, str]
        :returns: The created Attribute mapping
        :rtype: :class:`~tamr_unify_client.project.attribute_mapping.resource.AttributeMapping`
        """
        data = self.client.post(self.api_path, json=creation_spec).successful().json()
        return AttributeMapping(self.client, data)

    def delete_by_resource_id(self, resource_id):
        """Delete an attribute mapping using its Resource ID.

        :param resource_id: the resource ID of the mapping to be deleted.
        :type resource_id: str
        :returns: HTTP response from the server
        :rtype: :class:`requests.Response`
         """
        path = self.api_path + "/" + resource_id
        response = self.client.delete(path).successful()
        return response



from tamr_client._types import Instance, Session


def origin(instance: Instance) -> str:
    """HTTP origin i.e. :code:`<protocol>://<host>[:<port>]`.

    For additional information, see `MDN web docs <https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Origin>`_ .
    """
    if instance.port is None:
        return f"{instance.protocol}://{instance.host}"
    else:
        return f"{instance.protocol}://{instance.host}:{instance.port}"


def version(session: Session, instance: Instance) -> str:
    """Return the Tamr version for an instance.

    Args:
        session: Tamr Session
        instance: Tamr instance

    Returns: Version

    """
    # Version endpoints are not themselves versioned by design, but they are stable so they are ok to use here.
    return session.get(f"{origin(instance)}/api/versioned/service/version").json()[
        "version"
    ]



import requests

from tamr_client._types import Session


def from_auth(auth: requests.auth.HTTPBasicAuth) -> Session:
    """Create a new authenticated session

    Args:
        auth: Authentication
    """
    s = requests.Session()
    s.auth = auth
    return s



"""
See https://docs.tamr.com/reference#attribute-types
"""
import logging

from tamr_client._types import (
    Array,
    AttributeType,
    JsonDict,
    Map,
    PrimitiveType,
    Record,
)
from tamr_client._types import (  # noqa: F401
    BOOLEAN,
    DEFAULT,
    DOUBLE,
    GEOSPATIAL,
    INT,
    LONG,
    STRING,
)
from tamr_client.attribute import sub

logger = logging.getLogger(__name__)


def from_json(data: JsonDict) -> AttributeType:
    """Make an attribute type from JSON data (deserialize)

    Args:
        data: JSON data from Tamr server
    """
    base_type = data.get("baseType")
    if base_type is None:
        logger.error(f"JSON data: {repr(data)}")
        raise ValueError("Missing required field 'baseType'.")

    for primitive in PrimitiveType:
        if base_type == primitive.name:
            return primitive

    if base_type == Array._tag:
        inner_type = data.get("innerType")
        if inner_type is None:
            logger.error(f"JSON data: {repr(data)}")
            raise ValueError("Missing required field 'innerType' for Array type.")
        return Array(inner_type=from_json(inner_type))
    elif base_type == Map._tag:
        inner_type = data.get("innerType")
        if inner_type is None:
            logger.error(f"JSON data: {repr(data)}")
            raise ValueError("Missing required field 'innerType' for Map type.")
        return Map(inner_type=from_json(inner_type))
    elif base_type == Record._tag:
        attributes = data.get("attributes")
        if attributes is None:
            logger.error(f"JSON data: {repr(data)}")
            raise ValueError("Missing required field 'attributes' for Record type.")
        return Record(attributes=tuple([sub.from_json(attr) for attr in attributes]))
    else:
        logger.error(f"JSON data: {repr(data)}")
        raise ValueError(f"Unrecognized 'baseType': {base_type}")


def to_json(attr_type: AttributeType) -> JsonDict:
    """Serialize attribute type to JSON

    Args:
        attr_type: Attribute type to serialize
    """
    if isinstance(attr_type, PrimitiveType):
        return {"baseType": attr_type.name}
    elif isinstance(attr_type, (Array, Map)):
        return {
            "baseType": type(attr_type)._tag,
            "innerType": to_json(attr_type.inner_type),
        }
    elif isinstance(attr_type, Record):

        return {
            "baseType": type(attr_type)._tag,
            "attributes": [sub.to_json(attr) for attr in attr_type.attributes],
        }
    else:
        raise TypeError(attr_type)



from copy import deepcopy

from tamr_unify_client.base_resource import BaseResource
from tamr_unify_client.dataset.collection import DatasetCollection
from tamr_unify_client.dataset.resource import Dataset
from tamr_unify_client.project.attribute_configuration.collection import (
    AttributeConfigurationCollection,
)
from tamr_unify_client.project.attribute_mapping.collection import (
    AttributeMappingCollection,
)


class Project(BaseResource):
    """A Tamr project."""

    @classmethod
    def from_json(cls, client, resource_json, api_path=None):
        return super().from_data(client, resource_json, api_path)

    @property
    def name(self):
        """:type: str"""
        return self._data.get("name")

    @property
    def external_id(self):
        """:type: str"""
        return self._data.get("externalId")

    @property
    def description(self):
        """:type: str"""
        return self._data.get("description")

    @property
    def type(self):
        """A Tamr project type, listed in https://docs.tamr.com/reference#create-a-project.

        :type: str
        """
        return self._data.get("type")

    @property
    def attributes(self):
        """Attributes of this project.

        :return: Attributes of this project.
        :rtype: :class:`~tamr_unify_client.attribute.collection.AttributeCollection`
        """
        from tamr_unify_client.attribute.collection import AttributeCollection

        alias = self.api_path + "/attributes"
        return AttributeCollection(self.client, alias)

    def unified_dataset(self):
        """Unified dataset for this project.

        :return: Unified dataset for this project.
        :rtype: :class:`~tamr_unify_client.dataset.resource.Dataset`
        """
        alias = self.api_path + "/unifiedDataset"
        resource_json = self.client.get(alias).successful().json()
        return Dataset.from_json(self.client, resource_json, alias)

    def as_categorization(self):
        """Convert this project to a :class:`~tamr_unify_client.categorization.project.CategorizationProject`

        :return: This project.
        :rtype: :class:`~tamr_unify_client.categorization.project.CategorizationProject`
        :raises TypeError: If the :attr:`~tamr_unify_client.project.resource.Project.type` of this project is not ``"CATEGORIZATION"``
        """
        from tamr_unify_client.categorization.project import CategorizationProject

        if self.type != "CATEGORIZATION":
            raise TypeError(
                f"Cannot convert project to categorization project. Project type: {self.type}"
            )
        return CategorizationProject(self.client, self._data, self.api_path)

    def as_mastering(self):
        """Convert this project to a :class:`~tamr_unify_client.mastering.project.MasteringProject`

        :return: This project.
        :rtype: :class:`~tamr_unify_client.mastering.project.MasteringProject`
        :raises TypeError: If the :attr:`~tamr_unify_client.project.resource.Project.type` of this project is not ``"DEDUP"``
        """
        from tamr_unify_client.mastering.project import MasteringProject

        if self.type != "DEDUP":
            raise TypeError(
                f"Cannot convert project to mastering project. Project type: {self.type}"
            )
        return MasteringProject(self.client, self._data, self.api_path)

    def add_input_dataset(self, dataset):
        """
        Associate a dataset with a project in Tamr.

        By default, datasets are not associated with any projects.
        They need to be added as input to a project before they can be used
        as part of that project

        :param dataset: The dataset to associate with the project.
        :type dataset: :class:`~tamr_unify_client.dataset.resource.Dataset`
        :return: HTTP response from the server
        :rtype: :class:`requests.Response`
        """
        params = {"id": dataset.relative_id}
        response = self.client.post(
            self.api_path + "/inputDatasets", params=params
        ).successful()
        return response

    def remove_input_dataset(self, dataset):
        """Remove a dataset from a project.

        :param dataset: The dataset to be removed from this project.
        :type dataset: :class:`~tamr_unify_client.dataset.resource.Dataset`
        :return: HTTP response from the server
        :rtype: :class:`requests.Response`
        """
        params = {"id": dataset.relative_id}
        response = self.client.delete(
            self.api_path + "/inputDatasets", params=params
        ).successful()
        return response

    def input_datasets(self):
        """Retrieve a collection of this project's input datasets.

        :return: The project's input datasets.
        :rtype: :class:`~tamr_unify_client.dataset.collection.DatasetCollection`
        """
        alias = self.api_path + "/inputDatasets"
        return DatasetCollection(self.client, alias)

    def attribute_configurations(self):
        """Project's attribute's configurations.

        :returns: The configurations of the attributes of a project.
        :rtype: :class:`~tamr_unify_client.project.attribute_configuration.collection.AttributeConfigurationCollection`
        """
        alias = self.api_path + "/attributeConfigurations"
        info = AttributeConfigurationCollection(self.client, api_path=alias)
        return info

    def attribute_mappings(self):
        """Project's attribute's mappings.

       :returns: The attribute mappings of a project.
       :rtype: :class:`~tamr_unify_client.project.attribute_mapping.collection.AttributeMappingCollection`
       """
        alias = self.api_path + "/attributeMappings"
        info = AttributeMappingCollection(self.client, alias)
        return info

    def spec(self):
        """Returns this project's spec.

        :return: The spec for the project.
        :rtype: :class:`~tamr_unify_client.project.resource.ProjectSpec`
        """
        return ProjectSpec.of(self)

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"relative_id={self.relative_id!r}, "
            f"name={self.name!r}, "
            f"type={self.type!r})"
        )


class ProjectSpec:
    """A representation of the server view of a project."""

    def __init__(self, client, data, api_path):
        self.client = client
        self._data = data
        self.api_path = api_path

    @staticmethod
    def of(resource):
        """Creates a project spec from a project.

        :param resource: The existing project.
        :type resource: :class:`~tamr_unify_client.project.resource.Project`
        :return: The corresponding project spec.
        :rtype: :class:`~tamr_unify_client.project.resource.ProjectSpec`
        """
        return ProjectSpec(resource.client, deepcopy(resource._data), resource.api_path)

    @staticmethod
    def new():
        """Creates a blank spec that could be used to construct a new project.

        :return: The empty spec.
        :rtype: :class:`~tamr_unify_client.project.resource.ProjectSpec`
        """
        return ProjectSpec(None, {}, None)

    def from_data(self, data):
        """Creates a spec with the same client and API path as this one, but new data.

        :param data: The data for the new spec.
        :type data: dict
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.project.resource.ProjectSpec`
        """
        return ProjectSpec(self.client, data, self.api_path)

    def to_dict(self):
        """Returns a version of this spec that conforms to the API representation.

        :returns: The spec's dict.
        :rtype: dict
        """
        return deepcopy(self._data)

    def with_name(self, new_name):
        """Creates a new spec with the same properties, updating name.

        :param new_name: The new name.
        :type new_name: str
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.project.resource.ProjectSpec`
        """
        return self.from_data({**self._data, "name": new_name})

    def with_description(self, new_description):
        """Creates a new spec with the same properties, updating description.

        :param new_description: The new description.
        :type new_description: str
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.project.resource.ProjectSpec`
        """
        return self.from_data({**self._data, "description": new_description})

    def with_type(self, new_type):
        """Creates a new spec with the same properties, updating type.

        :param new_type: The new type.
        :type new_type: str
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.project.resource.ProjectSpec`
        """
        return self.from_data({**self._data, "type": new_type})

    def with_external_id(self, new_external_id):
        """Creates a new spec with the same properties, updating external ID.

        :param new_external_id: The new external ID.
        :type new_external_id: str
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.project.resource.ProjectSpec`
        """
        return self.from_data({**self._data, "externalId": new_external_id})

    def with_unified_dataset_name(self, new_unified_dataset_name):
        """Creates a new spec with the same properties, updating unified dataset name.

        :param new_unified_dataset_name: The new unified dataset name.
        :type new_unified_dataset_name: str
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.project.resource.ProjectSpec`
        """
        return self.from_data(
            {**self._data, "unifiedDatasetName": new_unified_dataset_name}
        )

    def put(self):
        """Commits these changes by updating the project in Tamr.

        :return: The updated project.
        :rtype: :class:`~tamr_unify_client.project.resource.Project`
        """
        updated_json = (
            self.client.put(self.api_path, json=self._data).successful().json()
        )
        return Project.from_json(self.client, updated_json, self.api_path)

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"dict={self._data})"
        )



from tamr_client._types.attribute import (
    Array,
    Attribute,
    AttributeType,
    BOOLEAN,
    ComplexType,
    DEFAULT,
    DOUBLE,
    GEOSPATIAL,
    INT,
    LONG,
    Map,
    PrimitiveType,
    Record,
    STRING,
    SubAttribute,
)
from tamr_client._types.auth import UsernamePasswordAuth
from tamr_client._types.backup import Backup
from tamr_client._types.dataset import AnyDataset, Dataset, UnifiedDataset
from tamr_client._types.instance import Instance
from tamr_client._types.json import JsonDict
from tamr_client._types.operation import Operation
from tamr_client._types.project import (
    CategorizationProject,
    GoldenRecordsProject,
    MasteringProject,
    Project,
    SchemaMappingProject,
)
from tamr_client._types.restore import Restore
from tamr_client._types.session import Session
from tamr_client._types.transformations import InputTransformation, Transformations
from tamr_client._types.url import URL



from __future__ import absolute_import

from base64 import b64encode

from requests.auth import HTTPBasicAuth
from requests.utils import to_native_string


def _basic_auth_str(username, password):
    auth = f"{username}:{password}"
    encoded = b64encode(auth.encode("latin1"))
    return "BasicCreds " + to_native_string(encoded.strip())


class UsernamePasswordAuth(HTTPBasicAuth):
    """Provides username/password authentication for Tamr.
    Specifically, sets the `Authorization` HTTP header with Tamr's custom `BasicCreds` format.

    :param str username:
    :param str password:

    Usage:
        >>> from tamr_unify_client.auth import UsernamePasswordAuth
        >>> auth = UsernamePasswordAuth('my username', 'my password')
        >>> import tamr_unify_client as api
        >>> unify = api.Client(auth)
    """

    def __init__(self, username, password):
        super(self.__class__, self).__init__(username, password)

    def __call__(self, r):
        r.headers["Authorization"] = _basic_auth_str(self.username, self.password)
        return r

    def __repr__(self):
        # intentionally leave out password (potentially sensitive)
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"username={self.username!r})"
        )



from dataclasses import dataclass, field
from typing import List

from tamr_client._types import Dataset


@dataclass(frozen=True)
class InputTransformation:
    transformation: str
    datasets: List[Dataset] = field(default_factory=list)


@dataclass(frozen=True)
class Transformations:
    input_scope: List[InputTransformation] = field(default_factory=list)
    unified_scope: List[str] = field(default_factory=list)



from tamr_client import response, Restore
from tamr_client._types import Instance, JsonDict, Session, URL
from tamr_client.exception import TamrClientException


class InvalidOperation(TamrClientException):
    """Raised when attempting an invalid operation.
    """

    pass


class NotFound(TamrClientException):
    """Raised when referencing a restore that does not exist on the server.
    """

    pass


def _from_json(url: URL, data: JsonDict) -> Restore:
    """Make restore from JSON data (deserialize).

    Args:
        url: Restore url
        data: Restore JSON data from Tamr server
    """
    return Restore(
        url=url,
        backup_path=data["backupPath"],
        state=data["state"],
        error_message=data["errorMessage"],
    )


def get(session: Session, instance: Instance) -> Restore:
    """Get information on the latest Tamr restore, if any.

    Args:
        session: Tamr session
        instance: Tamr instance

    Returns:
        Latest Tamr restore

    Raises:
        restore.NotFound: If no backup found at the specified URL
    """
    url = URL(instance=instance, path="instance/restore")
    r = session.get(str(url))
    if r.status_code == 404:
        raise NotFound(str(url))
    return _from_json(url, response.successful(r).json())


def initiate(session: Session, instance: Instance, backup_path: str) -> Restore:
    """Initiate a Tamr restore.

    Args:
        session: Tamr session
        instance: Tamr instance
        backup_path: Path to the backup

    Returns:
        Initiated restore

    Raises:
        restore.InvalidOperation: If attempting an invalid operation
    """
    url = URL(instance=instance, path="instance/restore")
    r = session.post(str(url), data=backup_path)
    if r.status_code == 400:
        raise InvalidOperation(str(url), r.json()["message"])
    return _from_json(url, response.successful(r).json())


def cancel(session: Session, restore: Restore) -> Restore:
    """Cancel a Tamr restore.

    Args:
        session: Tamr session
        restore: A Tamr restore

    Returns:
        Canceled restore

    Raises:
        restore.NotFound: If no backup file found at the specified path
        restore.InvalidOperation: If attempting an invalid operation
    """
    cancel_url = f"{restore.url}:cancel"
    r = session.post(cancel_url)
    if r.status_code == 404:
        raise NotFound(cancel_url)
    if r.status_code == 400:
        raise InvalidOperation(cancel_url, r.json()["message"])
    return _from_json(restore.url, response.successful(r).json())



from copy import deepcopy

from tamr_unify_client.base_resource import BaseResource


class AttributeConfiguration(BaseResource):
    """The configurations of Tamr Attributes.

   See https://docs.tamr.com/reference#the-attribute-configuration-object
   """

    @classmethod
    def from_json(
        cls, client, resource_json, api_path=None
    ) -> "AttributeConfiguration":
        return super().from_data(client, resource_json, api_path)

    @property
    def relative_id(self):
        """:type: str"""
        return self._data.get("relativeId")

    @property
    def id(self):
        """:type: str"""
        return self._data.get("id")

    @property
    def relative_attribute_id(self):
        """:type: str"""
        return self._data.get("relativeAttributeId")

    @property
    def attribute_role(self):
        """:type: str"""
        return self._data.get("attributeRole")

    @property
    def similarity_function(self):
        """:type: str"""
        return self._data.get("similarityFunction")

    @property
    def enabled_for_ml(self):
        """:type: bool"""
        return self._data.get("enabledForMl")

    @property
    def tokenizer(self):
        """:type: str"""
        return self._data.get("tokenizer")

    @property
    def numeric_field_resolution(self):
        """:type: list """
        return self._data.get("numericFieldResolution")

    @property
    def attribute_name(self):
        """:type: str"""
        return self._data.get("attributeName")

    def spec(self):
        """Returns this attribute configuration's spec.

         :return: The spec of this attribute configuration.
         :rtype: :class:`~tamr_unify_client.project.attribute_configuration.resource.AttributeConfigurationSpec`
         """
        return AttributeConfigurationSpec.of(self)

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"relative_id={self.relative_id!r}, "
            f"id={self.id!r}, "
            f"relative_attribute_id={self.relative_attribute_id!r}, "
            f"attribute_role={self.attribute_role!r}, "
            f"similarity_function={self.similarity_function!r}, "
            f"enabled_for_ml={self.enabled_for_ml!r}, "
            f"tokenizer={self.tokenizer!r}, "
            f"numeric_field_resolution={self.numeric_field_resolution!r}, "
            f"attribute_name={self.attribute_name!r})"
        )


class AttributeConfigurationSpec:
    """A representation of the server view of an attribute configuration."""

    def __init__(self, client, data, api_path):
        self.client = client
        self._data = data
        self.api_path = api_path

    @staticmethod
    def of(resource):
        """Creates an attribute configuration spec from an attribute configuration.

        :param resource: The existing attribute configuration.
        :type resource: :class:`~tamr_unify_client.project.attribute_configuration.resource.AttributeConfiguration`
        :return: The corresponding attribute creation spec.
        :rtype: :class:`~tamr_unify_client.project.attribute_configuration.resource.AttributeConfigurationSpec`
        """
        return AttributeConfigurationSpec(
            resource.client, deepcopy(resource._data), resource.api_path
        )

    @staticmethod
    def new():
        """Creates a blank spec that could be used to construct a new attribute configuration.

        :return: The empty spec.
        :rtype: :class:`~tamr_unify_client.project.attribute_configuration.resource.AttributeConfigurationSpec`
        """
        return AttributeConfigurationSpec(None, {}, None)

    def from_data(self, data):
        """Creates a spec with the same client and API path as this one, but new data.

        :param data: The data for the new spec.
        :type data: dict
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.project.attribute_configuration.resource.AttributeConfigurationSpec`
        """
        return AttributeConfigurationSpec(self.client, data, self.api_path)

    def to_dict(self):
        """Returns a version of this spec that conforms to the API representation.

        :returns: The spec's dict.
        :rtype: dict
        """
        return deepcopy(self._data)

    def with_attribute_role(self, new_attribute_role):
        """Creates a new spec with the same properties, updating attribute role.

        :param new_attribute_role: The new attribute role.
        :type new_attribute_role: str
        :return: A new spec.
        :rtype: :class:`~tamr_unify_client.project.attribute_configuration.resource.AttributeConfigurationSpec`
        """
        return self.from_data({**self._data, "attributeRole": new_attribute_role})

    def with_similarity_function(self, new_similarity_function):
        """Creates a new spec with the same properties, updating similarity function.

        :param new_similarity_function: The new similarity function.
        :type new_similarity_function: str
        :return: A new spec.
        :rtype: :class:`~tamr_unify_client.project.attribute_configuration.resource.AttributeConfigurationSpec`
        """
        return self.from_data(
            {**self._data, "similarityFunction": new_similarity_function}
        )

    def with_enabled_for_ml(self, new_enabled_for_ml):
        """Creates a new spec with the same properties, updating enabled for ML.

        :param new_enabled_for_ml: Whether the builder is enabled for ML.
        :type new_enabled_for_ml: bool
        :return: A new spec.
        :rtype: :class:`~tamr_unify_client.project.attribute_configuration.resource.AttributeConfigurationSpec`
        """
        return self.from_data({**self._data, "enabledForMl": new_enabled_for_ml})

    def with_tokenizer(self, new_tokenizer):
        """Creates a new spec with the same properties, updating tokenizer.

        :param new_tokenizer: The new tokenizer.
        :type new_tokenizer: str
        :return: A new spec.
        :rtype: :class:`~tamr_unify_client.project.attribute_configuration.resource.AttributeConfigurationSpec`
        """
        return self.from_data({**self._data, "tokenizer": new_tokenizer})

    def with_numeric_field_resolution(self, new_numeric_field_resolution):
        """Creates a new spec with the same properties, updating numeric field resolution.

        :param new_numeric_field_resolution: The new numeric field resolution.
        :type new_numeric_field_resolution: str
        :return: A new spec.
        :rtype: :class:`~tamr_unify_client.project.attribute_configuration.resource.AttributeConfigurationSpec`
        """
        return self.from_data(
            {**self._data, "numericFieldResolution": new_numeric_field_resolution}
        )

    def with_attribute_name(self, new_attribute_name):
        """Creates a new spec with the same properties, updating new attribute name.

        :param new_attribute_name: The new attribute name.
        :type new_attribute_name: str
        :return: A new spec.
        :rtype: :class:`~tamr_unify_client.project.attribute_configuration.resource.AttributeConfigurationSpec`
        """
        return self.from_data({**self._data, "attributeName": new_attribute_name})

    def put(self):
        """Updates the attribute configuration on the server.

        :return: The modified attribute configuration.
        :rtype: :class:`~tamr_unify_client.project.attribute_configuration.resource.AttributeConfiguration`
        """
        new_data = self.client.put(self.api_path, json=self._data).successful().json()
        return AttributeConfiguration.from_json(self.client, new_data, self.api_path)

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"dict={self._data})"
        )



"""
See https://docs.tamr.com/reference/dataset-models
"""
from copy import deepcopy
from dataclasses import replace
from typing import List, Optional, Tuple, Union

from tamr_client import operation, response
from tamr_client._types import (
    Attribute,
    Dataset,
    Instance,
    JsonDict,
    Operation,
    Session,
    URL,
)
from tamr_client.attribute import _from_json as _attribute_from_json
from tamr_client.exception import TamrClientException


class NotFound(TamrClientException):
    """Raised when referencing (e.g. updating or deleting) a dataset
    that does not exist on the server.
    """

    pass


class Ambiguous(TamrClientException):
    """Raised when referencing a dataset by name that matches multiple possible targets."""

    pass


class AlreadyExists(TamrClientException):
    """Raised when a dataset with these specifications already exists."""

    pass


def by_resource_id(session: Session, instance: Instance, id: str) -> Dataset:
    """Get dataset by resource ID

    Fetches dataset from Tamr server

    Args:
        instance: Tamr instance containing this dataset
        id: Dataset ID

    Raises:
        dataset.NotFound: If no dataset could be found at the specified URL.
            Corresponds to a 404 HTTP error.
        requests.HTTPError: If any other HTTP error is encountered.
    """
    url = URL(instance=instance, path=f"datasets/{id}")
    return _by_url(session, url)


def by_name(session: Session, instance: Instance, name: str) -> Dataset:
    """Get dataset by name

    Fetches dataset from Tamr server

    Args:
        instance: Tamr instance containing this dataset
        name: Dataset name

    Raises:
        dataset.NotFound: If no dataset could be found with that name.
        dataset.Ambiguous: If multiple targets match dataset name.
        requests.HTTPError: If any other HTTP error is encountered.
    """
    r = session.get(
        url=str(URL(instance=instance, path="datasets")),
        params={"filter": f"name=={name}"},
    )

    # Check that exactly one dataset is returned
    matches = r.json()
    if len(matches) == 0:
        raise NotFound(str(r.url))
    if len(matches) > 1:
        raise Ambiguous(str(r.url))

    # Make Dataset from response
    url = URL(instance=instance, path=matches[0]["relativeId"])
    return _from_json(url=url, data=matches[0])


def _by_url(session: Session, url: URL) -> Dataset:
    """Get dataset by URL

    Fetches dataset from Tamr server

    Args:
        url: Dataset URL

    Raises:
        dataset.NotFound: If no dataset could be found at the specified URL.
            Corresponds to a 404 HTTP error.
        requests.HTTPError: If any other HTTP error is encountered.
    """
    r = session.get(str(url))
    if r.status_code == 404:
        raise NotFound(str(url))
    data = response.successful(r).json()
    return _from_json(url, data)


def _from_json(url: URL, data: JsonDict) -> Dataset:
    """Make dataset from JSON data (deserialize)

    Args:
        url: Dataset URL
        data: Dataset JSON data from Tamr server
    """
    cp = deepcopy(data)
    return Dataset(
        url,
        name=cp["name"],
        description=cp.get("description"),
        key_attribute_names=tuple(cp["keyAttributeNames"]),
    )


def attributes(session: Session, dataset: Dataset) -> Tuple[Attribute, ...]:
    """Get all attributes from a dataset

    Args:
        dataset: Dataset containing the desired attributes

    Returns:
        The attributes for the specified dataset

    Raises:
        requests.HTTPError: If an HTTP error is encountered.
    """
    attrs_url = replace(dataset.url, path=dataset.url.path + "/attributes")
    r = session.get(str(attrs_url))
    attrs_json = response.successful(r).json()

    attrs = []
    for attr_json in attrs_json:
        id = attr_json["name"]
        attr_url = replace(attrs_url, path=attrs_url.path + f"/{id}")
        attr = _attribute_from_json(attr_url, attr_json)
        attrs.append(attr)
    return tuple(attrs)


def materialize(session: Session, dataset: Dataset) -> Operation:
    """Materialize a dataset and wait for the operation to complete
    Materializing consists of updating the dataset (including records) in persistent storage (HBase) based on upstream changes to data.

    Args:
        dataset: A Tamr dataset which will be materialized
    """
    op = _materialize_async(session, dataset)
    return operation.wait(session, op)


def _materialize_async(session: Session, dataset: Dataset) -> Operation:
    r = session.post(str(dataset.url) + ":refresh",)
    return operation._from_response(dataset.url.instance, r)


def delete(session: Session, dataset: Dataset, *, cascade: bool = False):
    """Deletes an existing dataset

    Sends a deletion request to the Tamr server

    Args:
        dataset: Existing dataset to delete
        cascade: Whether to delete all derived datasets as well

    Raises:
        dataset.NotFound: If no dataset could be found at the specified URL.
            Corresponds to a 404 HTTP error.
        requests.HTTPError: If any other HTTP error is encountered.
    """
    r = session.delete(str(dataset.url), params={"cascade": cascade})
    if r.status_code == 404:
        raise NotFound(str(dataset.url))
    response.successful(r)


def get_all(
    session: Session,
    instance: Instance,
    *,
    filter: Optional[Union[str, List[str]]] = None,
) -> Tuple[Dataset, ...]:
    """Get all datasets from an instance

    Args:
        instance: Tamr instance from which to get datasets
        filter: Filter expression, e.g. "externalId==wobbly"
            Multiple expressions can be passed as a list

    Returns:
        The datasets retrieved from the instance

    Raises:
        requests.HTTPError: If an HTTP error is encountered.
    """
    url = URL(instance=instance, path="datasets")

    if filter is not None:
        r = session.get(str(url), params={"filter": filter})
    else:
        r = session.get(str(url))

    datasets_json = response.successful(r).json()

    datasets = []
    for dataset_json in datasets_json:
        dataset_url = URL(instance=instance, path=dataset_json["relativeId"])
        dataset = _from_json(dataset_url, dataset_json)
        datasets.append(dataset)
    return tuple(datasets)


def create(
    session: Session,
    instance: Instance,
    *,
    name: str,
    key_attribute_names: Tuple[str, ...],
    description: Optional[str] = None,
    external_id: Optional[str] = None,
) -> Dataset:
    """Create a dataset in Tamr.

    Args:
        instance: Tamr instance
        name: Dataset name
        key_attribute_names: Dataset primary key attribute names
        description: Dataset description
        external_id: External ID of the dataset

    Returns:
        Dataset created in Tamr

    Raises:
        dataset.AlreadyExists: If a dataset with these specifications already exists.
        requests.HTTPError: If any other HTTP error is encountered.
    """
    data = {
        "name": name,
        "keyAttributeNames": key_attribute_names,
        "description": description,
        "externalId": external_id,
    }

    dataset_url = URL(instance=instance, path="datasets")
    r = session.post(url=str(dataset_url), json=data)

    if r.status_code == 400 and "already exists" in r.json()["message"]:
        raise AlreadyExists(r.json()["message"])

    data = response.successful(r).json()
    dataset_path = data["relativeId"]
    dataset_url = URL(instance=instance, path=str(dataset_path))

    return _by_url(session=session, url=dataset_url)



from tamr_unify_client.base_model import MachineLearningModel
from tamr_unify_client.categorization.taxonomy import Taxonomy
from tamr_unify_client.project.resource import Project


class CategorizationProject(Project):
    """A Categorization project in Tamr."""

    def model(self):
        """Machine learning model for this Categorization project.
        Learns from verified labels and predicts categorization labels for unlabeled records.

        :returns: The machine learning model for categorization.
        :rtype: :class:`~tamr_unify_client.base_model.MachineLearningModel`
        """
        alias = self.api_path + "/categorizations/model"
        return MachineLearningModel(self.client, None, alias)

    def create_taxonomy(self, creation_spec):
        """Creates a :class:`~tamr_unify_client.categorization.taxonomy.Taxonomy` for this project.

        A taxonomy cannot already be associated with this project.

        :param creation_spec: The creation specification for the taxonomy, which can include name.
        :type creation_spec: dict
        :returns: The new Taxonomy
        :rtype: :class:`~tamr_unify_client.categorization.taxonomy.Taxonomy`
        """
        alias = self.api_path + "/taxonomy"
        resource_json = self.client.post(alias, json=creation_spec).successful().json()
        return Taxonomy.from_json(self.client, resource_json, alias)

    def taxonomy(self):
        """Retrieves the :class:`~tamr_unify_client.categorization.taxonomy.Taxonomy` associated with this project.
        If a taxonomy is not already associated with this project,
        call :func:`~tamr_unify_client.categorization.project.CategorizationProject.create_taxonomy` first.

        :returns: The project's Taxonomy
        :rtype: :class:`~tamr_unify_client.categorization.taxonomy.Taxonomy`
        """
        alias = self.api_path + "/taxonomy"
        resource_json = self.client.get(alias).successful().json()
        return Taxonomy.from_json(self.client, resource_json, alias)

    # super.__repr__ is sufficient



from tamr_unify_client.base_resource import BaseResource
from tamr_unify_client.operation import Operation


class DatasetProfile(BaseResource):
    """Profile info of a Tamr dataset."""

    @classmethod
    def from_json(cls, client, resource_json, api_path=None) -> "DatasetProfile":
        return super().from_data(client, resource_json, api_path)

    @property
    def dataset_name(self) -> str:
        """The name of the associated dataset.

        :type: str
        """
        return self._data.get("datasetName")

    @property
    def relative_dataset_id(self) -> str:
        """The relative dataset ID of the associated dataset.

        :type: str
        """
        return self._data.get("relativeDatasetId")

    @property
    def is_up_to_date(self) -> bool:
        """Whether the associated dataset is up to date.

        :type: bool
        """
        return self._data.get("isUpToDate")

    @property
    def profiled_data_version(self) -> str:
        """The profiled data version.

        :type: str
        """
        return self._data.get("profiledDataVersion")

    @property
    def profiled_at(self) -> dict:
        """Info about when profile info was generated.

        :type: dict
        """
        return self._data.get("profiledAt")

    @property
    def simple_metrics(self) -> list:
        """Simple metrics for profiled dataset.

        :type: list
        """
        return self._data.get("simpleMetrics")

    @property
    def attribute_profiles(self) -> list:
        """Simple metrics for profiled dataset.

        :type: list
        """
        return self._data.get("attributeProfiles")

    def refresh(self, **options):
        """Updates the dataset profile if needed.

        The dataset profile is updated on the server; you will need to call
        :func:`~tamr_unify_client.dataset.resource.Dataset.profile`
        to retrieve the updated profile.

        :param ``**options``: Options passed to underlying :class:`~tamr_unify_client.operation.Operation` .
            See :func:`~tamr_unify_client.operation.Operation.apply_options` .
        :returns: The refresh operation.
        :rtype: :class:`~tamr_unify_client.operation.Operation`
        """
        response = self.client.post(self.api_path + ":refresh").successful()
        op = Operation.from_response(self.client, response)
        return op.apply_options(**options)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"relative_id={self.relative_id!r}, "
            f"dataset_name={self.dataset_name!r}, "
            f"relative_dataset_id={self.relative_dataset_id!r}, "
            f"is_up_to_date={self.is_up_to_date!r}, "
            f"profiled_data_version={self.profiled_data_version!r}, "
            f"profiled_at={self.profiled_at!r}, "
            f"simple_metrics={self.simple_metrics!r})"
        )



from tamr_unify_client.mastering.published_cluster.record_version import (
    RecordPublishedClusterVersion,
)


class RecordPublishedCluster:
    """A representation of a published cluster of a record in a mastering project with version information.
    See https://docs.tamr.com/reference#retrieve-published-clusters-given-record-ids.

    This is not a `BaseResource` because it does not have its own API endpoint.

    :param data: The JSON entity representing this
        :class:`~tamr_unify_client.mastering.published_cluster.record.RecordPublishedCluster`.
    """

    def __init__(self, data):
        self._data = data

    @property
    def entity_id(self):
        """:type: str"""
        return self._data.get("entityId")

    @property
    def source_id(self):
        """:type: str"""
        return self._data.get("sourceId")

    @property
    def origin_entity_id(self):
        """:type: str"""
        return self._data.get("originEntityId")

    @property
    def origin_source_id(self):
        """:type: str"""
        return self._data.get("originSourceId")

    @property
    def versions(self):
        """:type: list[:class:`~tamr_unify_client.mastering.published_cluster.record_version.RecordPublishedClusterVersion`]"""
        return [RecordPublishedClusterVersion(v) for v in self._data.get("versions")]

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"entity_id={self.entity_id!r}, "
            f"source_id={self.source_id!r}, "
            f"origin_entity_id={self.origin_entity_id!r}, "
            f"origin_source_id={self.origin_source_id!r})"
        )



from tamr_unify_client.mastering.published_cluster.version import (
    PublishedClusterVersion,
)


class PublishedCluster:
    """A representation of a published cluster in a mastering project with version information.
    See https://docs.tamr.com/reference#retrieve-published-clusters-given-cluster-ids.

    This is not a `BaseResource` because it does not have its own API endpoint.

    :param data: The JSON entity representing this :class:`~tamr_unify_client.mastering.published_cluster.resource.PublishedCluster`.
    """

    def __init__(self, data):
        self._data = data

    @property
    def id(self):
        """:type: str"""
        return self._data.get("id")

    @property
    def versions(self):
        """:type: list[:class:`~tamr_unify_client.mastering.published_cluster.version.PublishedClusterVersion`]"""
        return [PublishedClusterVersion(v) for v in self._data.get("versions")]

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"id={self.id!r})"
        )



# Logging
#########

import logging

# https://docs.python-guide.org/writing/logging/#logging-in-a-library
logging.getLogger(__name__).addHandler(logging.NullHandler())

# BETA check
############

from tamr_client import _beta

_beta.check()

# types
#######

from tamr_client._types import (
    AnyDataset,
    Attribute,
    AttributeType,
    Backup,
    CategorizationProject,
    Dataset,
    GoldenRecordsProject,
    InputTransformation,
    Instance,
    MasteringProject,
    Operation,
    Project,
    Restore,
    SchemaMappingProject,
    Session,
    SubAttribute,
    Transformations,
    UnifiedDataset,
    URL,
    UsernamePasswordAuth,
)

# functionality
###############

from tamr_client import attribute
from tamr_client import backup
from tamr_client import categorization
from tamr_client import dataset
from tamr_client import golden_records
from tamr_client import instance
from tamr_client import mastering
from tamr_client import operation
from tamr_client import primary_key
from tamr_client import project
from tamr_client import response
from tamr_client import restore
from tamr_client import schema_mapping
from tamr_client import session
from tamr_client import transformations
from tamr_client.dataset import dataframe
from tamr_client.dataset import record
from tamr_client.exception import TamrClientException



from copy import deepcopy


class AttributeMapping:
    """see https://docs.tamr.com/reference#retrieve-projects-mappings
    AttributeMapping and AttributeMappingCollection do not inherit from BaseResource and BaseCollection.
    BC and BR require a specific URL for each individual attribute mapping
    (ex: /projects/1/attributeMappings/1), but these types of URLs do not exist for attribute mappings
    """

    def __init__(self, client, data):
        self._data = data
        self.client = client
        # AttributeMapping cannot be aliased, and Project cannot be aliased,
        # so AttributeMapping only ever has one address, which is both
        # its relative_id and its api_path.
        self.api_path = self.relative_id

    @property
    def id(self):
        """:type: str"""
        return self._data["id"]

    @property
    def relative_id(self):
        """:type: str"""
        return self._data["relativeId"]

    @property
    def input_attribute_id(self):
        """:type: str"""
        return self._data["inputAttributeId"]

    @property
    def relative_input_attribute_id(self):
        """:type: str"""
        return self._data["relativeInputAttributeId"]

    @property
    def input_dataset_name(self):
        """:type: str"""
        return self._data["inputDatasetName"]

    @property
    def input_attribute_name(self):
        """:type: str"""
        return self._data["inputAttributeName"]

    @property
    def unified_attribute_id(self):
        """:type: str"""
        return self._data["unifiedAttributeId"]

    @property
    def relative_unified_attribute_id(self):
        """:type: str"""
        return self._data["relativeUnifiedAttributeId"]

    @property
    def unified_dataset_name(self):
        """:type: str"""
        return self._data["unifiedDatasetName"]

    @property
    def unified_attribute_name(self):
        """:type: str"""
        return self._data["unifiedAttributeName"]

    @property
    def resource_id(self):
        """:type: str"""
        spliced = self.relative_id.split("attributeMappings/")[1]
        return spliced

    def spec(self):
        """Returns a spec representation of this attribute mapping.

        :return: The attribute mapping spec.
        :rtype: :class:`~tamr_unify_client.project.attribute_mapping.resource.AttributeMappingSpec`
        """
        return AttributeMappingSpec.of(self)

    def delete(self):
        """Delete this attribute mapping.

        :return: HTTP response from the server
        :rtype: :class:`requests.Response`
        """
        response = self.client.delete(self.api_path).successful()
        return response

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"id={self.id!r}, "
            f"relative_id={self.relative_id!r}, "
            f"input_attribute_id={self.input_attribute_id!r}, "
            f"relative_input_attribute_id={self.relative_input_attribute_id!r}, "
            f"input_dataset_name={self.input_dataset_name!r}, "
            f"input_attribute_name={self.input_attribute_name!r}, "
            f"unified_attribute_id={self.unified_attribute_id!r}, "
            f"relative_unified_attribute_id={self.relative_unified_attribute_id!r}, "
            f"unified_dataset_name={self.unified_dataset_name!r}, "
            f"unified_attribute_name={self.unified_attribute_name!r})"
        )


class AttributeMappingSpec:
    """A representation of the server view of an attribute mapping"""

    def __init__(self, data):
        self._data = data

    @staticmethod
    def of(resource):
        """Creates an attribute mapping spec from a attribute mapping.

        :param resource: The existing attribute mapping.
        :type resource: :class:`~tamr_unify_client.project.attribute_mapping.resource.AttributeMapping`
        :return: The corresponding attribute mapping spec.
        :rtype: :class:`~tamr_unify_client.project.attribute_mapping.resource.AttributeMappingSpec`
        """
        return AttributeMappingSpec(deepcopy(resource._data))

    @staticmethod
    def new():
        """Creates a blank spec that could be used to construct a new attribute mapping.

        :return: The empty spec.
        :rtype: :class:`~tamr_unify_client.project.attribute_mapping.resource.AttributeMappingSpec`
        """
        return AttributeMappingSpec({})

    def to_dict(self):
        """Returns a version of this spec that conforms to the API representation.

        :returns: The spec's dict.
        :rtype: dict
        """
        return deepcopy(self._data)

    def with_input_attribute_id(self, new_input_attribute_id):
        """Creates a new spec with the same properties, updating the input attribute id.

        :param new_input_attribute_id: The new input attribute id.
        :type new_input_attribute_id: str
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.project.attribute_mapping.resource.AttributeMappingSpec`
        """
        return AttributeMappingSpec(
            {**self._data, "inputAttributeId": new_input_attribute_id}
        )

    def with_relative_input_attribute_id(self, new_relative_input_attribute_id):
        """Creates a new spec with the same properties, updating the relative input attribute id.

        :param new_relative_input_attribute_id: The new relative input attribute Id.
        :type new_relative_input_attribute_id: str
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.project.attribute_mapping.resource.AttributeMappingSpec`
        """
        return AttributeMappingSpec(
            {**self._data, "relativeInputAttributeId": new_relative_input_attribute_id}
        )

    def with_input_dataset_name(self, new_input_dataset_name):
        """Creates a new spec with the same properties, updating the input dataset name.

        :param new_input_dataset_name: The new input dataset name.
        :type new_input_dataset_name: str
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.project.attribute_mapping.resource.AttributeMappingSpec`
        """
        return AttributeMappingSpec(
            {**self._data, "inputDatasetName": new_input_dataset_name}
        )

    def with_input_attribute_name(self, new_input_attribute_name):
        """Creates a new spec with the same properties, updating the input attribute name.

        :param new_input_attribute_name: The new input attribute name.
        :type new_input_attribute_name: str
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.project.attribute_mapping.resource.AttributeMappingSpec`
        """
        return AttributeMappingSpec(
            {**self._data, "inputAttributeName": new_input_attribute_name}
        )

    def with_unified_attribute_id(self, new_unified_attribute_id):
        """Creates a new spec with the same properties, updating the unified attribute id.

        :param new_unified_attribute_id: The new unified attribute id.
        :type new_unified_attribute_id: str
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.project.attribute_mapping.resource.AttributeMappingSpec`
        """
        return AttributeMappingSpec(
            {**self._data, "unifiedAttributeId": new_unified_attribute_id}
        )

    def with_relative_unified_attribute_id(self, new_relative_unified_attribute_id):
        """Creates a new spec with the same properties, updating the relative unified attribute id.

        :param new_relative_unified_attribute_id: The new relative unified attribute id.
        :type new_relative_unified_attribute_id: str
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.project.attribute_mapping.resource.AttributeMappingSpec`
        """
        return AttributeMappingSpec(
            {
                **self._data,
                "relativeUnifiedAttributeId": new_relative_unified_attribute_id,
            }
        )

    def with_unified_dataset_name(self, new_unified_dataset_name):
        """Creates a new spec with the same properties, updating the unified dataset name.

        :param new_unified_dataset_name: The new unified dataset name.
        :type new_unified_dataset_name: str
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.project.attribute_mapping.resource.AttributeMappingSpec`
        """
        return AttributeMappingSpec(
            {**self._data, "unifiedDatasetName": new_unified_dataset_name}
        )

    def with_unified_attribute_name(self, new_unified_attribute_name):
        """Creates a new spec with the same properties, updating the unified attribute name.

        :param new_unified_attribute_name: The new unified attribute name.
        :type new_unified_attribute_name: str
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.project.attribute_mapping.resource.AttributeMappingSpec`
        """
        return AttributeMappingSpec(
            {**self._data, "unifiedAttributeName": new_unified_attribute_name}
        )

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"dict={self._data})"
        )



from requests.auth import AuthBase


class TokenAuth(AuthBase):
    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers["Authorization"] = "BasicCreds " + self.token
        return r

    def __repr__(self):
        # intentionally leave out the token (potentially sensitive)
        return f"{self.__class__.__module__}." f"{self.__class__.__qualname__}()"



from tamr_unify_client.base_collection import BaseCollection
from tamr_unify_client.project.resource import Project


class ProjectCollection(BaseCollection):
    """Collection of :class:`~tamr_unify_client.project.resource.Project` s.

    :param client: Client for API call delegation.
    :type client: :class:`~tamr_unify_client.Client`
    :param api_path: API path used to access this collection.
        Default: ``"projects"``.
    :type api_path: str
    """

    def __init__(self, client, api_path="projects"):
        super().__init__(client, api_path)

    def by_resource_id(self, resource_id):
        """Retrieve a project by resource ID.

        :param resource_id: The resource ID. E.g. ``"1"``
        :type resource_id: str
        :returns: The specified project.
        :rtype: :class:`~tamr_unify_client.project.resource.Project`
        """
        return super().by_resource_id("projects", resource_id)

    def by_relative_id(self, relative_id):
        """Retrieve a project by relative ID.

        :param relative_id: The resource ID. E.g. ``"projects/1"``
        :type relative_id: str
        :returns: The specified project.
        :rtype: :class:`~tamr_unify_client.project.resource.Project`
        """
        return super().by_relative_id(Project, relative_id)

    def by_external_id(self, external_id):
        """Retrieve a project by external ID.

        :param external_id: The external ID.
        :type external_id: str
        :returns: The specified project, if found.
        :rtype: :class:`~tamr_unify_client.project.resource.Project`
        :raises KeyError: If no project with the specified external_id is found
        :raises LookupError: If multiple projects with the specified external_id are found
        """
        return super().by_external_id(Project, external_id)

    def stream(self):
        """Stream projects in this collection. Implicitly called when iterating
        over this collection.

        :returns: Stream of projects.
        :rtype: Python generator yielding :class:`~tamr_unify_client.project.resource.Project`

        Usage:
            >>> for project in collection.stream(): # explicit
            >>>     do_stuff(project)
            >>> for project in collection: # implicit
            >>>     do_stuff(project)
        """
        return super().stream(Project)

    def by_name(self, project_name: str) -> Project:
        """Get project by name

        Fetches a specific project in this collection by exact-match on name.

        Args:
            project_name: Name of the desired project.
        Raises:
            KeyError: If no project with specified name was found.
        """
        for project in self:
            if project.name == project_name:
                return project
        raise KeyError(f"No project found with name: {project_name}")

    def create(self, creation_spec):
        """
        Create a Project in Tamr

        :param creation_spec: Project creation specification should be formatted as specified in the `Public Docs for Creating a Project <https://docs.tamr.com/reference#create-a-project>`_.
        :type creation_spec: dict[str, str]
        :returns: The created Project
        :rtype: :class:`~tamr_unify_client.project.resource.Project`
        """
        data = self.client.post(self.api_path, json=creation_spec).successful().json()
        return Project.from_json(self.client, data)

    # super.__repr__ is sufficient



from tamr_unify_client.base_resource import BaseResource


class DatasetStatus(BaseResource):
    """Streamability status of a Tamr dataset."""

    @classmethod
    def from_json(cls, client, resource_json, api_path=None) -> "DatasetStatus":
        return super().from_data(client, resource_json, api_path)

    @property
    def dataset_name(self) -> str:
        """The name of the associated dataset.

        :type: str
        """
        return self._data.get("datasetName")

    @property
    def relative_dataset_id(self) -> str:
        """The relative dataset ID of the associated dataset.

        :type: str
        """
        return self._data.get("relativeDatasetId")

    @property
    def is_streamable(self) -> bool:
        """Whether the associated dataset is available to be streamed.

        :type: bool
        """
        return self._data.get("isStreamable")

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"relative_id={self.relative_id!r}, "
            f"dataset_name={self.dataset_name!r}, "
            f"relative_dataset_id={self.relative_dataset_id!r}, "
            f"is_streamable={self.is_streamable!r})"
        )



"""
Convenient functionality for interacting with pandas DataFrames.
"""

import json
import os
from typing import Optional, TYPE_CHECKING

from tamr_client import primary_key
from tamr_client._types import Dataset, JsonDict, Session
from tamr_client.dataset import record

BUILDING_DOCS = os.environ.get("TAMR_CLIENT_DOCS") == "1"
if TYPE_CHECKING or BUILDING_DOCS:
    import pandas as pd


def upsert(
    session: Session,
    dataset: Dataset,
    df: "pd.DataFrame",
    *,
    primary_key_name: Optional[str] = None,
) -> JsonDict:
    """Upserts a record for each row of `df` with attributes for each column in `df`.

    Args:
        dataset: Dataset to receive record updates
        df: The DataFrame containing records to be upserted
        primary_key_name: The primary key of the dataset.  Must be a column of `df`. By default the key_attribute_name of dataset

    Returns:
        JSON response body from the server

    Raises:
        requests.HTTPError: If an HTTP error is encountered
        primary_key.NotFound: If `primary_key_name` is not a column in `df` or the index of `df`
        ValueError: If `primary_key_name` matches both a column in `df` and the index of `df`
    """
    if primary_key_name is None:
        primary_key_name = dataset.key_attribute_names[0]

    # preconditions
    if primary_key_name in df.columns and primary_key_name == df.index.name:
        raise primary_key.Ambiguous(
            f"Index {primary_key_name} has the same name as column {primary_key_name}"
        )
    elif primary_key_name not in df.columns and primary_key_name != df.index.name:
        raise primary_key.NotFound(
            f"Primary key: {primary_key_name} is not DataFrame index name: {df.index.name} or in DataFrame column names: {df.columns}"
        )

    # promote primary key column to index
    if primary_key_name in df.columns:
        df = df.set_index(primary_key_name)

    # serialize records via to_json to handle `np.nan` values
    serialized_records = ((pk, row.to_json()) for pk, row in df.iterrows())
    records = (
        {primary_key_name: pk, **json.loads(row)} for pk, row in serialized_records
    )
    return record.upsert(session, dataset, records, primary_key_name=primary_key_name)
